/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */
package org.objectweb.medor.datasource.lib;

import org.objectweb.medor.api.EvaluationException;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.datasource.api.Wrapper;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.VariableOperand;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.lib.BasicVariableOperand;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.TCQueryLeaf;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.tuple.lib.MemoryTuple;

public class TCWrapper implements Wrapper {

    public TupleCollection fetchData(QueryLeaf ql, ParameterOperand[] parameters, Object conn)
            throws EvaluationException {
        return new FilteredTC((TCQueryLeaf) ql, parameters);
    }

    class FilteredTC implements TupleCollection {

        private TCQueryLeaf tcQueryLeaf;
        private Expression filter;
        private TupleCollection subTupleCollection;
        private ParameterOperand[] parameters;
        private Tuple tupleBuffer1,tupleBuffer2;
        private VariableOperand[] operandBuffer1,operandBuffer2;
        private int cursor;
        private boolean empty, isTheLast = false, isBuffer2Free = false;
        Operand filterResult;
		private boolean closed = false;

        public FilteredTC(TCQueryLeaf tcQueryLeaf, ParameterOperand[] parameters)
                throws EvaluationException {
            this.tcQueryLeaf = tcQueryLeaf;
            this.parameters = parameters;
            this.filter = tcQueryLeaf.getQueryFilter();
			try {
				this.filterResult = filter.compileExpression();
			}
			catch (ExpressionException e) {
				throw new EvaluationException(e);
			}

			// Init the two buffers
            try {
                operandBuffer1 = new BasicVariableOperand[tcQueryLeaf.getTupleStructure().getSize()];
                operandBuffer2 = new BasicVariableOperand[tcQueryLeaf.getTupleStructure().getSize()];
                for (int cpt = 0; (cpt < tcQueryLeaf.getTupleStructure().getSize()); cpt++) {
                    operandBuffer1[cpt] = new BasicVariableOperand(tcQueryLeaf.getTupleStructure().getField(cpt + 1).getType());
                    operandBuffer2[cpt] = new BasicVariableOperand(tcQueryLeaf.getTupleStructure().getField(cpt + 1).getType());
                }

                tupleBuffer1 = new MemoryTuple(operandBuffer1); // Creating the Buffer1
                tupleBuffer2 = new MemoryTuple(operandBuffer2);
                subTupleCollection = tcQueryLeaf.getTupleCollection();
                empty = (subTupleCollection.isEmpty());
                if (empty)
                    cursor = 0;
                else
                    this.init();
            } catch (MedorException mexp) {
                throw new EvaluationException(mexp);
            }
        }

		public void close() throws MedorException {
			closed = true;
		}

        public TupleStructure getMetaData() throws MedorException {
            return tcQueryLeaf.getTupleStructure();
        }

        public boolean isLast() throws MedorException {
            return isTheLast;
        }

        public int card() throws MedorException {
            return cursor;
        }

        public boolean next() throws MedorException {
            Tuple currentTuple = null;
            boolean moved = false;

            if (isEmpty() || (cursor == -2))
                moved = false;
            else if (isLast()) {
                // we have iterated all tuples
                cursor = -2;
                moved = false;
            } else {
				try {
					// we move to the next Result
					moved = true;
					cursor++;
					do {
						//Verify if there is others results
						if (subTupleCollection.next()) {
							currentTuple = subTupleCollection.getTuple();
							filter.evaluate(parameters, currentTuple);
						} else {
							isTheLast = true;
						}
					} while (!(isTheLast) && !(filterResult.getBoolean()));
					if (!isTheLast) {
						if (isBuffer2Free) {
							//Load the next Result in the buffer2
							tcQueryLeaf.getTupleLoader().loadTuple(currentTuple, operandBuffer2, parameters);
							isBuffer2Free = false;
						} else {
							//Load the next Result in the buffer1
							tcQueryLeaf.getTupleLoader().loadTuple(currentTuple, operandBuffer1, parameters);
							isBuffer2Free = true;
						}
					} else
						isBuffer2Free = !isBuffer2Free;
				}
				catch (ExpressionException e) {
					throw new MedorException(e);
				}
			}
            return moved;
        }


        /**
         * This method Move the cursor to the first result if there is one, load it, then it verfy if it is the last one.
         * In the case of the last one it returns false whether it returns true.
         */
        private boolean init() throws MedorException {
            isTheLast = false;
            Tuple currentTuple;
            isBuffer2Free = false;
            subTupleCollection.first();
            currentTuple = subTupleCollection.getTuple();
			try {
				filter.evaluate(parameters, currentTuple);
				while (!empty && !filterResult.getBoolean()) {
					if (subTupleCollection.next()) {
						currentTuple = subTupleCollection.getTuple();
						filter.evaluate(parameters, currentTuple);
					} else {
						empty = true;//This TupleCollection result is empty
					}
				}
				if (empty) {
					cursor = -2;
				} else {
					// Loading the first result in the buffer1
					tcQueryLeaf.getTupleLoader().loadTuple(currentTuple, operandBuffer1, parameters);
					cursor = 1;
					do {
						// Verfy if there is other result
						if (subTupleCollection.next()) {
							currentTuple = subTupleCollection.getTuple();
							filter.evaluate(parameters, currentTuple);
						} else {
							// There is no more results
							isTheLast = true;
						}
					} while (!(isTheLast) && !(filterResult.getBoolean()));
					if (!isTheLast) {
						// Loading the next Result in the buffer2
						tcQueryLeaf.getTupleLoader().loadTuple(currentTuple, operandBuffer2, parameters);
					}
				}
			}
			catch (ExpressionException e) {
				throw new MedorException(e);
			}
			return (!empty);
        }

        public void first() throws MedorException {
            //The Initialisation
            init();
        }

        public int getRow() throws MedorException {
            return cursor;
        }

        public Tuple getTuple() throws MedorException {
            if (!isEmpty() && getRow() >= 1) {
                if (isBuffer2Free)
                    return tupleBuffer2;
                else
                    return tupleBuffer1;
            } else
                throw new MedorException(" No elements fetched in this TupleCollection");
        }

        public synchronized Tuple getTuple(int numTuple) throws MedorException {
            row(numTuple);
            return getTuple();
        }

        public boolean isEmpty() throws MedorException {
            return empty;
        }

        public boolean row(int numTuple) throws MedorException {
            first();
            int cpt = 1;
            boolean stop = false;
            while ((cpt < numTuple) && !stop) {
                if (!next())
                    stop = true;
                else
                    cpt++;
            }
            if (cpt < numTuple)
                return false;
            else {
                cursor = numTuple;
                return true;
            }
        }
    }
}
