/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */

package org.objectweb.medor.datasource.rdb.lib;

import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapter;
import org.objectweb.jorm.mapper.rdb.adapter.BasicRdbAdapter;
import org.objectweb.jorm.mapper.rdb.lib.RdbConnectionWrapper;
import org.objectweb.jorm.api.PException;
import org.objectweb.medor.api.EvaluationException;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.datasource.api.Wrapper;
import org.objectweb.medor.expression.api.CalculatedParameterOperand;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.filter.lib.InCollection;
import org.objectweb.medor.lib.Log;
import org.objectweb.medor.query.api.FilteredQueryTree;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.rdb.api.RdbExpQueryLeaf;
import org.objectweb.medor.query.rdb.api.RdbQueryLeaf;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.tuple.lib.EmptyTupleCollection;
import org.objectweb.medor.type.lib.QTypeTuple;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.io.Serializable;
import java.io.IOException;
import java.math.BigDecimal;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Date;
import java.util.Collection;
import java.util.Iterator;

/**
 * This class implements the Wrapper interface for the jdbc resources. This
 * implementation uses Statements to perform the request.
 *
 * @author Mourad Alia
 */
public class JDBCWrapper implements Wrapper {

	/**
	 * this field references a Logger which is never null.
	 */
	static Logger log = null;
	static Logger TClog = null;
	boolean debug;

	public JDBCWrapper() {
		if (log == null) {
			log = Log.loggerFactory.getLogger(
				Log.MEDOR_PREFIX + ".eval.rdb.JDBCWrapper");
			TClog = Log.loggerFactory.getLogger(
				Log.MEDOR_PREFIX + ".eval.rdb.JDBCTC");
		}
	}

	public TupleCollection fetchData(QueryLeaf lf,
									 ParameterOperand[] parameters,
									 Object connection)
			throws EvaluationException {
		debug = log.isLoggable(BasicLevel.DEBUG);
		if (debug) {
			log.log(BasicLevel.DEBUG, "Parameters: " + posToString(parameters));
		}
		RdbQueryLeaf leaf = (RdbQueryLeaf) lf;
		ParameterOperand[] usedParameters = null;
		//Calculate the parameter list of the leaf with the parameter list of the query
		RdbAdapter rdbAdapt = getRdbAdapter(leaf);
		Connection conn = getConnection(connection);
		ResultSet resultSet = null;
		PreparedStatement ps = null;
		String sqlRequest;
		synchronized(leaf) {
			if (parameters != null
					&& parameters.length > 0
					&& leaf instanceof FilteredQueryTree) {
				Expression filter = ((FilteredQueryTree) leaf).getQueryFilter();
				if (filter != null) {
					ArrayList al = getUsedParameters(filter, parameters, null);
					if (al != null && al.size() >0) {
						usedParameters = (ParameterOperand[])
								al.toArray(new ParameterOperand[al.size()]);
					}
				}
			}
			if (debug) {
				log.log(BasicLevel.DEBUG, "Used parameters: " + posToString(parameters));
			}
			try {
				sqlRequest = leaf.getSqlRequest(usedParameters);
				if (debug) {
					log.log(BasicLevel.DEBUG, "SQL query for leaf " + leaf + ": " + sqlRequest);
				}
				ps = conn.prepareStatement(
						sqlRequest,
						ResultSet.TYPE_SCROLL_INSENSITIVE,
						ResultSet.CONCUR_READ_ONLY);
				if (usedParameters != null) {
					assignParameter(ps, usedParameters, rdbAdapt);
				}
			} catch (Exception e) {
				throw new EvaluationException(e);
			}
		}
		if (debug) {
			log.log(BasicLevel.DEBUG, "Evaluate the query: " + sqlRequest);
		}
		try {
			resultSet = ps.executeQuery();
			if (!resultSet.next()) {
				if (debug) log.log(BasicLevel.DEBUG, "No result, closing the ResultSet and the PreparedStatement");
				resultSet.close();
				ps.close();
				return new EmptyTupleCollection(lf.getTupleStructure());
			} else {
				if (debug) log.log(BasicLevel.DEBUG, "There is a result");
				return new JDBCTupleCollection(
						lf.getTupleStructure(), resultSet, ps, rdbAdapt, TClog);
			}
		} catch (EvaluationException ee) {
			throw ee;
		} catch (Exception mexp) {
			throw new EvaluationException(mexp);
		}
	}

	private Connection getConnection(Object connection) throws EvaluationException {
		try {
			return RdbConnectionWrapper.narrow2SQL(connection);
		} catch (PException e) {
			throw new EvaluationException(
					"Impossible to evaluate a jdbc leaf:", e);
		}
	}

	private RdbAdapter getRdbAdapter(RdbQueryLeaf leaf) {
		RdbAdapter rdbAdapt = null;
		if (leaf instanceof RdbExpQueryLeaf) {
			rdbAdapt = ((RdbExpQueryLeaf) leaf).getRdbAdapter();
		}
		if (rdbAdapt == null) {
			rdbAdapt = new BasicRdbAdapter();
		}
		return rdbAdapt;
	}

	private void assignParameter(PreparedStatement ps,
								 ParameterOperand[] usedParameters,
								 RdbAdapter rdbAdapt)
			throws MedorException, ExpressionException, SQLException, IOException {
		for (int i = 0; i < usedParameters.length; i++) {
			ParameterOperand po = usedParameters[i];
			if (debug) {
				log.log(BasicLevel.DEBUG, "Assign parameter (name="
			        + po.getName() + ", value=" + po + ")");
			}
			switch (po.getType().getTypeCode()) {
			case QTypeTuple.TYPECODE_INT:
				rdbAdapt.setInt(ps, i + 1, po.getInt());
				break;
			case QTypeTuple.TYPECODE_OBJINT:
				rdbAdapt.setOint(ps, i + 1, (Integer) po.getObject());
				break;
			case QTypeTuple.TYPECODE_SHORT:
				rdbAdapt.setShort(ps, i + 1, po.getShort());
				break;
			case QTypeTuple.TYPECODE_OBJSHORT:
				rdbAdapt.setOshort(ps, i + 1, (Short) po.getObject());
				break;
			case QTypeTuple.TYPECODE_BYTE:
				rdbAdapt.setByte(ps, i + 1, po.getByte());
				break;
			case QTypeTuple.TYPECODE_OBJBYTE:
				rdbAdapt.setObyte(ps, i + 1, (Byte) po.getObject());
				break;
			case QTypeTuple.TYPECODE_LONG:
				rdbAdapt.setLong(ps, i + 1, po.getLong());
				break;
			case QTypeTuple.TYPECODE_OBJLONG:
				rdbAdapt.setOlong(ps, i + 1, (Long) po.getObject());
				break;
			case QTypeTuple.TYPECODE_DOUBLE:
				rdbAdapt.setDouble(ps, i + 1, po.getDouble());
				break;
			case QTypeTuple.TYPECODE_OBJDOUBLE:
				rdbAdapt.setOdouble(ps, i + 1, (Double) po.getObject());
				break;
			case QTypeTuple.TYPECODE_BOOLEAN:
				rdbAdapt.setBoolean(ps, i + 1, po.getBoolean());
				break;
			case QTypeTuple.TYPECODE_OBJBOOLEAN:
				rdbAdapt.setOboolean(ps, i + 1, (Boolean) po.getObject());
				break;
			case QTypeTuple.TYPECODE_FLOAT:
				rdbAdapt.setFloat(ps, i + 1, po.getFloat());
				break;
			case QTypeTuple.TYPECODE_OBJFLOAT:
				rdbAdapt.setOfloat(ps, i + 1, (Float) po.getObject());
				break;
			case QTypeTuple.TYPECODE_BIGDECIMAL:
			case QTypeTuple.TYPECODE_BIGINTEGER:
				rdbAdapt.setBigDecimal(ps, i + 1, (BigDecimal) po.getObject());
				break;
			case QTypeTuple.TYPECODE_STRING:
				rdbAdapt.setString(ps, i + 1, (String) po.getObject());
				break;
			case QTypeTuple.TYPECODE_CHAR:
				rdbAdapt.setChar(ps, i + 1, po.getChar());
				break;
			case QTypeTuple.TYPECODE_OBJCHAR:
				rdbAdapt.setOchar(ps, i + 1, (Character) po.getObject());
				break;
			case QTypeTuple.TYPECODE_SERIALIZED:
				rdbAdapt.setSerialized(ps, i + 1, (Serializable) po.getObject());
				break;
			case QTypeTuple.TYPECODE_DATE:
				rdbAdapt.setDate(ps, i + 1, (Date) po.getObject(), null);
				break;
			case QTypeTuple.TYPECODE_BYTEARRAY:
				rdbAdapt.setByteArray(ps, i + 1, (byte[]) po.getObject());
				break;
			case QTypeTuple.TYPECODE_CHARARRAY:
				rdbAdapt.setCharArray(ps, i + 1, (char[]) po.getObject());
				break;
			case QTypeTuple.TYPECODE_COLLECTION:
				//Already assign in the sql request building
				break;
			default:
				throw new EvaluationException("The type "
						+ po.getType().getJormName()
						+ " is not supported '"
						+ po.getType().getTypeCode() + "'"
				);
			}
		}
	}

	private ArrayList getUsedParameters(Expression e,
										ParameterOperand[] parameters,
										ArrayList al)
			throws EvaluationException {
		//TODO take memberof into account for parameters
		if ((e instanceof InCollection) &&
				!(((Operator)e).getExpression(1) instanceof ParameterOperand)
			) {
			al = getUsedParameters(((Operator) e).getExpression(0),
					parameters, al);
			al = getUsedParametersInCollection(
					(Collection) ((Operand)((Operator) e).getExpression(1)).getObject(),
					parameters,
					al);
		} else if (e instanceof Operator) {
			for (int i = 0; i < ((Operator) e).getOperandNumber(); i++) {
				al = getUsedParameters(((Operator) e).getExpression(i),
						parameters, al);
			}
		} else if (e instanceof CalculatedParameterOperand) {
			CalculatedParameterOperand cpo = (CalculatedParameterOperand) e;
			try {
				cpo.evaluate(parameters);
			} catch (ExpressionException e1) {
				throw new EvaluationException(e1);
			}
			if (al == null)
				al = new ArrayList();
			al.add(cpo);
		} else if (e instanceof ParameterOperand) {
			int i = 0;
			ParameterOperand param = (ParameterOperand) e;
			while (i < parameters.length
					&& !parameters[i].getName().equals(param.getName()))
				i++;
			if (i < parameters.length) {
				if (parameters[i].getType().getTypeCode() !=
						param.getType().getTypeCode()) {
					throw new EvaluationException("Bad parameter type: found "
							+ parameters[i].getType().getJormName()
							+ " / expected " + param.getType().getJormName());
				}
				if (al == null)
					al = new ArrayList();
				al.add(parameters[i]);
			} else {
				StringBuffer msg = new StringBuffer();
				msg.append("You must specify the value of parameter ");
				msg.append(param.getName());
				msg.append(". Found: {");
				String sep = "";
				for (int j = 0; j < parameters.length; j++) {
					msg.append(sep);
					msg.append(parameters[j].getName());
					sep = ", ";
				}
				msg.append("}");
				throw new EvaluationException(msg.toString());
			}
		}
		return al;

	}

	private ArrayList getUsedParametersInCollection(
			Collection col,
			ParameterOperand[] parameters,
			ArrayList al)
			throws EvaluationException {
		Iterator colit = col.iterator();
		while (colit.hasNext()) {
			Object e = colit.next();
			if (e instanceof ParameterOperand) {
				int i = 0;
				ParameterOperand param = (ParameterOperand) e;
				while (i < parameters.length
						&& !parameters[i].getName().equals(param.getName()))
					i++;
				if (i < parameters.length) {
					if (parameters[i].getType().getTypeCode() !=
							param.getType().getTypeCode()) {
						throw new EvaluationException("Bad parameter type: found "
								+ parameters[i].getType().getJormName()
								+ " / expected " + param.getType().getJormName());
					}
					if (al == null)
						al = new ArrayList();
					al.add(parameters[i]);
				} else {
					String msg = "You must specify the value of parameter "
							+ param.getName() + ". Found: {";
					String sep = "";
					for (int j = 0; j < parameters.length; j++) {
						msg += sep + parameters[j].getName();
						sep = ", ";
					}
					msg += "}";
					throw new EvaluationException(msg);
				}
			}
		}
		return al;
	}

	private String posToString(ParameterOperand[] pos) {
		if (pos == null) {
			return "null";
		}
		StringBuffer sb = new StringBuffer("[");
		String sep = "";
		for(int i=0; i<pos.length; i++) {
			sb.append(sep);
			sb.append("(name='");
			sb.append("'");
			sb.append(pos[i].getName());
			sb.append("', value=");
			sb.append(pos[i]);
			sb.append(")");
		}
		sb.append("]");
		return sb.toString();
	}
}
