/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.eval.api;

import org.objectweb.medor.api.EvaluationException;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.datasource.api.WrapperFactory;
import org.objectweb.medor.eval.prefetch.api.PrefetchBuffer;

/**
 * This interface represents the MEDOR evaluator. It permits to evaluate an
 * evaluable (optimized) query Tree.
 *
 */
public interface QueryEvaluator {

    /**
     * Launch the evaluation of a query.
     * @param parameters is the array of query parameter (Never null).
     * @param ressources is a ConnectionResources containing associations between
     * the DataStore names used in the QueryTree and the connection objects to
     * use for the evaluation.It can be null.
     * @param parameters is a ConnectionResources containing associations between
     * the DataStore names used in the QueryTree and the connection objects to
     * use for the evaluation.
     */
    TupleCollection evaluate(ParameterOperand[] parameters,
                             ConnectionResources ressources)
            throws EvaluationException;
    /**
     * Launch the evaluation of a query. This method is able to prefetch data
     * of the query in a prefetch buffer.
     * @param parameters is the array of query parameter (Never null).
     * @param ressources is a ConnectionResources containing associations between
     * the DataStore names used in the QueryTree and the connection objects to
     * use for the evaluation.It can be null.
     * @param pb is the prefetch buffer which must be fill with query results.
     * If this parameter is null then no data prefetching is done.
     */
    TupleCollection evaluate(ParameterOperand[] parameters,
                             ConnectionResources ressources,
                             PrefetchBuffer pb)
            throws EvaluationException;

    ConnectionResources getRequiredConnectionResources();

    /**
     * @return the cache size used in different node of the QueryNode tree.
     */
    long getCacheSize();

    /**
     * @return the evaluable (optimized) query tree.
     */
    QueryTree getQueryTree();

    /**
     * Returns the associated WrapperFactory. This is used to be able to
     * bind specific Wrappers to DataStore types.
     * @return the associated WrapperFactory
     */
    WrapperFactory getWrapperFactory();
}
