/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

/**
 * Package definition.
 */

package org.objectweb.medor.eval.lib;

import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.eval.api.BinaryEvaluatedTC;
import org.objectweb.medor.eval.api.NodeEvaluator;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.VariableOperand;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.lib.BasicVariableOperand;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.tuple.lib.MemoryTuple;

/**
 * This class represents TupleCollection result of filtering or joining then
 * projecting of two sub TupleCollection.
 * <p>It is typically used for performing
 * JoinProject and Intersection Operations. The indexes of the sub
 * TupleCollection are getted by the to methods: getRightTCCursor() and
 * getLeftTCCursor(). The type of the evaluation used is the Iterative one.
 */
public class BinaryImplicitTC
		extends BasicBinaryEvalutedTC
		implements BinaryEvaluatedTC {

    private ParameterOperand[] parameters;
    private QueryNode query;
    private Expression filter;
    private TupleCollection leftTCResult;
    private TupleCollection rightTCResult;
    private Tuple tupleBuffer1,tupleBuffer2,concatResult;
    private VariableOperand[] operandBuffer1,operandBuffer2,attConcatBuffers;
    private int cursor;
    private int leftTCSize, rightTCSize;
    private boolean empty, isBuffer2Free = false ,isTheLast = false;
    private Operand filterResult;

    public BinaryImplicitTC(QueryNode query,
                            NodeEvaluator leftNodeEvaluator,
                            NodeEvaluator rightNodeEvaluator,
                            ParameterOperand[] parameters)
		throws MedorException {
		this.closed = false;
        this.parameters = parameters;
        this.query = query;
        this.filter = query.getQueryFilter();
		try {
			this.filterResult = filter.compileExpression();
		}
		catch (ExpressionException e) {
			throw new MedorException(e);
		}

		// Fetching the left TupleCollection
        leftTCResult = leftNodeEvaluator.fetchData(parameters);

        // Fetching the right TupleCollection
        rightTCResult = rightNodeEvaluator.fetchData(parameters);

        leftTCSize = leftTCResult.getMetaData().getSize();
        rightTCSize = rightTCResult.getMetaData().getSize();

        // init the buffers of results
        operandBuffer1 = new VariableOperand[query.getTupleStructure().getSize()];
        operandBuffer2 = new VariableOperand[query.getTupleStructure().getSize()];
        for (int cpt = 0; (cpt < query.getTupleStructure().getSize()); cpt++) {
            operandBuffer1[cpt] =
				new BasicVariableOperand(
					query.getTupleStructure().getField(cpt + 1).getType());
            operandBuffer2[cpt] =
				new BasicVariableOperand(
					query.getTupleStructure().getField(cpt + 1).getType());
        }
        tupleBuffer1 = new MemoryTuple(operandBuffer1);// Creating the buffer Tuple
        tupleBuffer2 = new MemoryTuple(operandBuffer2);

        // Init the Tuple concat
        attConcatBuffers = new VariableOperand[leftTCSize + rightTCSize];
        for (int cpt = 0; (cpt < leftTCSize); cpt++) {
            attConcatBuffers[cpt] =
				new BasicVariableOperand(
					leftTCResult.getMetaData().getField(cpt + 1).getType());
        }

        for (int cpt = leftTCSize; (cpt < leftTCSize + rightTCSize); cpt++) {
            attConcatBuffers[cpt] =
				new BasicVariableOperand(
					rightTCResult.getMetaData().getField(cpt + 1 - leftTCSize).getType());
        }
        concatResult = new MemoryTuple(attConcatBuffers); // Creating the buffer Tuple

        // One method evaluation is to supose that if one of the two TupleCollection is empty then this TupleCollection
        // object is then empty.
        empty = (leftTCResult.isEmpty() || rightTCResult.isEmpty());
        if (empty) {
            cursor = 0;
        } else {
            // move the cursor to the first result if there is one and verfy if it is the last one
			try {
				init();
			}
			catch (ExpressionException e) {
				throw new MedorException(e);
			}
		}
    }

    public TupleStructure getMetaData() throws MedorException {
        return query.getTupleStructure();
    }

    public boolean isLast() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        return isTheLast;
    }

	public void close() throws MedorException {
		closed = true;
		if (leftTCResult!=null)
			leftTCResult.close();
		if (rightTCResult!=null)
			rightTCResult.close();
	}

    public int card() throws MedorException {
        return cursor;
    }

    public synchronized boolean next() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        Tuple leftTuple;
        Tuple rightTuple;
        boolean isMoved = false;

        if (isEmpty() || (cursor == -2))
            isMoved = false;
        else if (isLast()) {
            // All the tuples are iterated. we are at the last one
            cursor = -2;
            isMoved = false;
        } else {
			try {
				isMoved = true; // we are really at the next Tuple result
				cursor++;
				do {
					// We verify if this tuple is the last one or not.
					if (rightTCResult.next()) {
						leftTuple = leftTCResult.getTuple();
						rightTuple = rightTCResult.getTuple();
						concatTuple(leftTuple, rightTuple);
						filter.evaluate(parameters, concatResult);
					} else {
						if (leftTCResult.next()) {
							rightTCResult.first();
							rightTuple = rightTCResult.getTuple();
							leftTuple = leftTCResult.getTuple();
							// the result of the concatenation is always in the resultTuple
							concatTuple(leftTuple, rightTuple);
							filter.evaluate(parameters, concatResult);
						} else
						// There no tuples to fetch
							isTheLast = true;
					}
				} while (!(isTheLast) && !(filterResult.getBoolean()));
				if (!isTheLast) {
					if (isBuffer2Free) {
						// Loading the next result and storing in the nextTuple buffer
						query.getTupleLoader().loadTuple(
							concatResult, operandBuffer2, parameters);
						isBuffer2Free = false;
					} else {
						// Loading the next result and storing in the currentTuple buffer
						query.getTupleLoader().loadTuple(
							concatResult, operandBuffer1, parameters);
						isBuffer2Free = true;
					}

				} else {
					// Switch loadong result to the free buffer
					isBuffer2Free = !isBuffer2Free;
				}
			}
			catch (ExpressionException e) {
				throw new MedorException(e);
			}
		}
        return isMoved;
    }

    /**
     * This method Move the cursor to the first result if there is one, load it, then it verfy if it is the last one.
     * In the case of the last one it returns false whether it returns true.
     */
    private boolean init() throws MedorException, ExpressionException {

        leftTCResult.first();// go to the first Tuple to iterate
        rightTCResult.first();
        isTheLast = false;
        Tuple leftTuple = leftTCResult.getTuple();// Get the left tuple
        Tuple rightTuple = rightTCResult.getTuple(); // Get the right Tuple
        concatTuple(leftTuple, rightTuple);
        // Start iterating and filtering to get the first Result
        filter.evaluate(parameters, concatResult);
        while (!empty && !filterResult.getBoolean()) {
            if (rightTCResult.next()) {
                rightTuple = rightTCResult.getTuple();
                concatTuple(leftTuple, rightTuple);
                filter.evaluate(parameters, concatResult);
            } else {
                if (leftTCResult.next()) {
                    rightTCResult.first();
                    rightTuple = rightTCResult.getTuple();
                    leftTuple = leftTCResult.getTuple();
                    concatTuple(leftTuple, rightTuple);
                    filter.evaluate(parameters, concatResult);
                } else//There is no result for this TupleCollection Object
                    empty = true;
            }
        }
        if (empty) {
            cursor = 0;
        } else {
            // Loading the first Result in the buffer1
            query.getTupleLoader().loadTuple(
				concatResult, operandBuffer1, parameters);

            //Init the cursor
            cursor = 1;
            do {
                // Verfy if this first Result is not the last one
                if (rightTCResult.next()) {
                    rightTuple = rightTCResult.getTuple();
                    concatTuple(leftTuple, rightTuple);
                    filter.evaluate(parameters, concatResult);
                } else {
                    if (leftTCResult.next()) {
                        rightTCResult.first();
                        rightTuple = rightTCResult.getTuple();
                        leftTuple = leftTCResult.getTuple();
                        concatTuple(leftTuple, rightTuple);
                        filter.evaluate(parameters, concatResult);
                    } else //There is only one result for this TupleCollection Object
                        isTheLast = true;
                }
            } while (!(isTheLast) && !(filterResult.getBoolean()));
            if (!isTheLast) {
                // Loading the next Result in the buffer2
                query.getTupleLoader().loadTuple(
					concatResult, operandBuffer2, parameters);
            }
        }
        return (!empty);
    }

    public synchronized void first() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        //The Initialization
		try {
			init();
		}
		catch (ExpressionException e) {
			e.printStackTrace();  //To change body of catch statement use Options | File Templates.
		}
	}

    public synchronized int getRow() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        return cursor;
    }

    public synchronized Tuple getTuple() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        if (!isEmpty() && getRow() >= 1) {
            if (isBuffer2Free)
                return tupleBuffer2;
            else
                return tupleBuffer1;
        } else
            throw new MedorException(" No elements fetched in this TupleCollection " + getRow());
    }

    public synchronized boolean isEmpty() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        return empty;
    }

    public synchronized boolean row(int numTuple) throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        if ((cursor > numTuple) || (cursor < 0)) {
            first();
        }
        boolean stop = false;
        while ((cursor < numTuple) && !stop) {
            if (!next()) {
                stop = true;
            }
        }
        if (cursor != numTuple)
            return false; // Tuple index error
        else {
            return true;
        }

    }

    public synchronized Tuple getTuple(int numTuple) throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        row(numTuple);
        return getTuple();
    }

    public synchronized int getLeftTCCursor() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        return leftTCResult.getRow();
    }

    public synchronized int getRightTCCursor() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        return rightTCResult.getRow();
    }

    private void concatTuple(Tuple t1, Tuple t2) throws MedorException {
        for (int cpt = 0; (cpt < leftTCSize); cpt++) {
            attConcatBuffers[cpt] =
				(VariableOperand) t1.getLikeOperand(cpt + 1);
        }
        for (int cpt = 0; (cpt < rightTCSize); cpt++) {
            attConcatBuffers[cpt + leftTCSize] =
				(VariableOperand) t2.getLikeOperand(cpt + 1);
        }
    }
}