/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.eval.lib;

import org.objectweb.medor.api.EvaluationException;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.datasource.api.Wrapper;
import org.objectweb.medor.datasource.api.WrapperFactory;
import org.objectweb.medor.datasource.lib.MedorWrapperFactory;
import org.objectweb.medor.eval.api.ConnectionResources;
import org.objectweb.medor.eval.api.NodeEvaluator;
import org.objectweb.medor.eval.api.NodeEvaluatorFactory;
import org.objectweb.medor.eval.api.QueryEvaluator;
import org.objectweb.medor.eval.prefetch.api.PrefetchBuffer;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.rdb.lib.AggregateRdbQueryNode;
import org.objectweb.medor.tuple.api.TupleCollection;

import java.util.HashMap;
import java.util.Stack;

/**
 * This class is a simple evaluator of query tree, supporting the data
 * prefetching
 *
 * @author S.Chassande-Barrioz
 */
public class MedorEvaluator implements QueryEvaluator {

    /**
     * The query tree which must be evaluated
     */
    private QueryTree query;

    /**
     * The query tree to evalued, stored as a stack
     */
    private Stack queryTreeStack;

    /**
     * The factory of wrappers (leaf evaluator)
     */
    private WrapperFactory wrapperFactory;

    /**
     * The factory of node evaluator (QueryNode evaluator)
     */
    private NodeEvaluatorFactory nodeEvaluatorFactory;

    /**
     * The cache size for each  node evaluator
     */
    private long cacheSize;

	public MedorEvaluator(QueryTree query, long cacheSize)
		throws EvaluationException {
		this.query = query;
		this.cacheSize = cacheSize; // The all cahe size
        nodeEvaluatorFactory = new IteratifNodeEvaluatorFactory();
        wrapperFactory = new MedorWrapperFactory();
        queryTreeStack = new Stack();
		toStack(query, queryTreeStack);
	}

    private void toStack(QueryTree query, Stack pile)
        throws EvaluationException {
        pile.push(query);
        if ((query instanceof QueryNode) &&
				!(query instanceof AggregateRdbQueryNode)) {
            QueryTree[] children = ((QueryNode) query).getChildren();
            for(int i=(children.length-1); i>=0; i--) {
                toStack(children[i], pile);
            }
        }
    }

    // IMPLEMENTATION OF THE QueryEvaluator INTERFACE //
    //------------------------------------------------//

	public WrapperFactory getWrapperFactory() {
		return wrapperFactory;
	}

    /**
	 * First, it allocates a tree of Evaluator (Wrapper or NodeEvaluator) for
     * the query tree associated to this evaluator. In a second step it launches
     * the evalution of the query.
     * @param parameters is the array of parameters
     * @param resources is the holder of connection used by the wrapper to
     * access the data support.
	 */
	public TupleCollection evaluate(ParameterOperand[] parameters,
									ConnectionResources resources)
		throws EvaluationException {
        return evaluate(parameters, resources, null);
    }

    /**
     * @param parameters is a ConnectionResources containing associations between
     * the DataStore names used in the QueryTree and the connection objects to
     * use for the evaluation.
     * @param prefetchBuffer is the prefetchBuffer instance where buffer data
     * must be put. If this field is null, then no data prefetching mus be done.
     */
    public TupleCollection evaluate(ParameterOperand[] parameters,
                                    ConnectionResources resources,
                                    PrefetchBuffer prefetchBuffer)
            throws EvaluationException {
		try {
			Stack runTimeStack = new Stack();
			for(int cpt = queryTreeStack.size() - 1; cpt >= 0; cpt--) {
				Object queryTree = queryTreeStack.elementAt(cpt);
				if (queryTree instanceof QueryLeaf) {
                    //allocate a wrapper for the query leaf
					QueryLeaf ql = (QueryLeaf) queryTree;
					Wrapper w = wrapperFactory.getWrapper(ql.getDataStore());
                    TupleCollection tc = w.fetchData(ql, parameters,
                            (resources != null
                            ? resources.getConnection(ql)
                            : null));
					runTimeStack.push(new TupleCollectionNodeEvaluator(tc));
				} else {
                    //allocate a node evaluator for the query node
					QueryNode qn = (QueryNode) queryTree;
					QueryTree[] children = qn.getChildren();
                    NodeEvaluator[] evals = new NodeEvaluator[children.length];
                    for(int i=0; i<children.length; i++) {
                        evals[i] = (NodeEvaluator) runTimeStack.pop();
                        evals[i].setCacheSize(cacheSize);
                    }
                    runTimeStack.push(nodeEvaluatorFactory
                            .createNodeEvaluator(qn, evals));
				}
			}
			NodeEvaluator queryTreeHead = ((NodeEvaluator) runTimeStack.pop());
			queryTreeHead.setCacheSize(cacheSize);
            queryTreeHead.setPrefetchBuffer(prefetchBuffer);
			return queryTreeHead.fetchData(parameters);
		}
		catch (MedorException mexp) {
			throw new EvaluationException(mexp);
		}
	}

	public QueryTree getQueryTree() {
		return query;
	}

	public long getCacheSize() {
		return cacheSize;
	}

	public ConnectionResources getRequiredConnectionResources() {
		HashMap qls = new HashMap(1);
        int size = queryTreeStack.size();
		for (int i = 0; i<size; i++) {
			if (queryTreeStack.elementAt(i) instanceof QueryLeaf) {
				qls.put(queryTreeStack.elementAt(i), null);
			}
		}
		return new MedorConnectionResources(qls);
	}
}
