/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

/**
 * Package definition.
 */

package org.objectweb.medor.eval.lib;

import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.eval.api.NodeEvaluator;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.VariableOperand;
import org.objectweb.medor.query.api.NestedField;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.UnnestQueryNode;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.tuple.lib.MemoryTuple;

/**
 * This class represent a TupleCollection Result of the Unnest Operation.
 * We don't store buffers in this operation because all Tuples of the sub
 * TupleCollection will be in this TupleCollection, there no filtering.
 */
public class UnnestEvaluatedTC implements TupleCollection {
	private TupleStructure tcStructure;
	private ParameterOperand[] parameters;
	private int sizeOfResult, sizeOfNotUnnestedFields, nestedFieldRank;
	private NestedField unnestField;
	private VariableOperand[] currentAttBuffers;
	private Tuple nestedCurrentTuple, currentBuffer;
	private TupleCollection nestedFieldTC, unnestedTupleCollection;
	private int cursor;
	private int[] indexes;
	private boolean closed = false;

	public UnnestEvaluatedTC(UnnestQueryNode unnestQueryTree,
							 NodeEvaluator subNodeEvaluator,
							 ParameterOperand[] parameters)
		throws MedorException {
		this.tcStructure = unnestQueryTree.getTupleStructure();
		this.parameters = parameters;

		// the Field wich will be degrouped
		unnestField = unnestQueryTree.getUnnestedField();

		// the sub TupleCollection to Unnest
		unnestedTupleCollection = subNodeEvaluator.fetchData(parameters);

		// the size of this TupleCollection
		sizeOfResult = unnestQueryTree.getTupleStructure().getSize();

		//the number of fields wich will be projected without the fields
		//of the nestedField to unnest
		sizeOfNotUnnestedFields = unnestQueryTree.getProjectedFields().length;

		/**
         * This int array contains indexes of the not projectd fields in the
		 * sub TupleCollection
         * This operation should be moved when constructing or compiling the
		 * Query
         */
		indexes = new int[sizeOfNotUnnestedFields];
		QueryTree[] childs = unnestQueryTree.getChildren();
		PropagatedField[] notNestedFields =
			unnestQueryTree.getProjectedFields();

		for (int cpt = 0; (cpt < sizeOfNotUnnestedFields); cpt++) {
			indexes[cpt] =
				childs[0].getTupleStructure().getFieldRank(
					notNestedFields[cpt].getPreviousFields()[0]);
		}
		// the rank of the field to unnest
		nestedFieldRank =
			childs[0].getTupleStructure().getFieldRank(unnestField);
		currentAttBuffers = new VariableOperand[sizeOfResult];
		currentBuffer = new MemoryTuple(currentAttBuffers);
		init();
	}

	private void init() throws MedorException {
		cursor = 1;
		if (!isEmpty()) {
			nestedCurrentTuple = unnestedTupleCollection.getTuple();
			nestedFieldTC = nestedCurrentTuple.getTupleCollection(
				indexes[nestedFieldRank]);
			for (int cpt = 0; (cpt < sizeOfNotUnnestedFields); cpt++) {
				currentAttBuffers[cpt] =
					(VariableOperand) nestedCurrentTuple.getLikeOperand(
						indexes[cpt]);
			}
			if (!(nestedFieldTC.isEmpty())) {
				for (int cpt = sizeOfNotUnnestedFields;
					 (cpt < sizeOfResult);
					 cpt++) {
					currentAttBuffers[cpt] =
						(VariableOperand)
						nestedFieldTC.getTuple().getLikeOperand(
							cpt - sizeOfNotUnnestedFields + 1);
				} // currentBuffer contain now the first result
			}
		}
	}

	public void close() throws MedorException {
		closed = true;
		if (nestedFieldTC != null) {
			nestedFieldTC.close();
		}
		if (unnestedTupleCollection != null) {
			unnestedTupleCollection.close();
		}
	}

	/**
	 * It contains metadata of fields wich will be projected
	 */
	public TupleStructure getMetaData() throws MedorException {
		return tcStructure;
	}

	public boolean isLast() throws MedorException {
		return ((unnestedTupleCollection.isLast()) && (nestedFieldTC.isLast()));
	}

	public boolean next() throws MedorException {
		if (nestedFieldTC.next()) {
			for (int cpt = sizeOfNotUnnestedFields;
				 (cpt < sizeOfResult);
				 cpt++) {
				currentAttBuffers[cpt] =
					(VariableOperand) nestedFieldTC.getTuple().getLikeOperand(
						cpt - sizeOfNotUnnestedFields + 1);
			}
			cursor++;
			return true;
		}
		else {
			if (unnestedTupleCollection.next()) {
				nestedCurrentTuple = unnestedTupleCollection.getTuple();
				nestedFieldTC =
					nestedCurrentTuple.getTupleCollection(nestedFieldRank);
				for (int cpt = 0; (cpt < sizeOfNotUnnestedFields); cpt++) {
					currentAttBuffers[cpt] =
						(VariableOperand) nestedCurrentTuple.getLikeOperand(
							indexes[cpt]);
				}

				// we supose that the nestedFieldTC is not empty
				if (!(nestedFieldTC.isEmpty())) {
					for (int cpt = sizeOfNotUnnestedFields;
						 (cpt < sizeOfResult);
						 cpt++) {
						currentAttBuffers[cpt] =
							(VariableOperand)
							nestedFieldTC.getTuple().getLikeOperand(
								cpt - sizeOfNotUnnestedFields + 1);
					}
				}
				cursor++;
				return true;
			}
			else
				return false;
		}
	}

	public void first() throws MedorException {
		init();
	}

	public int getRow() throws MedorException {
		return cursor;
	}

	public Tuple getTuple() throws MedorException {
		return currentBuffer;
	}

	public synchronized Tuple getTuple(int numTuple) throws MedorException {
		return getTuple();
	}

	public boolean row(int row) throws MedorException {
		return false;

	}

	public int card() throws MedorException {
		return 0;
	}

	public boolean isEmpty() throws MedorException {
		return unnestedTupleCollection.isEmpty();
	}
}
