/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre, S. Chassande-Barrioz
 */

package org.objectweb.medor.eval.prefetch.lib;

import org.objectweb.medor.eval.prefetch.api.PrefetchBuffer;
import org.objectweb.medor.eval.prefetch.api.PrefetchBufferHolder;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.api.MedorException;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.HashMap;
import java.util.Map;

/**
 * @author  P. Dechamboux
 */
public class PrefetchBufferImpl implements PrefetchBuffer {
	private int indexPos;
	private TupleCollection tupleCollection = null;
	private Map tupleMap = null;
	private Logger logger;
	private boolean multithread;
	private boolean closed;

	PrefetchBufferImpl(int indexpos, boolean multithread, Logger logger) {
		this.indexPos = indexpos;
		this.logger = logger;
		this.multithread = multithread;
		this.closed = false;
	}

	// IMPLEMENTATION OF METHODS FROM THE PrefetchBuffer INTERFACE

	public void addPrefetchTuple() throws MedorException {
		TupleCollection localTC = tupleCollection;
		if (localTC == null) {
			return;
		}
		try {
			int row = localTC.getRow();
			if (row <1) {
				return;
			}
			if (tupleMap == null) {
				tupleMap = new HashMap();
			}
			Tuple pt = localTC.getTuple();
			Object index = pt.getObject(indexPos);
			if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
				logger.log(BasicLevel.DEBUG, "Add Tuple in PreftechBuffer:"
						+ "\n\tindex=" + index
						+ "\n\trow=" + row);
			}
			tupleMap.put(index, new Integer(row));
		} catch (MedorException e) {
			if (logger != null) {
				logger.log(BasicLevel.WARN,
					"Error during the adding of a prefetched tuple: ", e);
			}
		}
	}

	public void setTupleCollection(TupleCollection tc) throws MedorException {
		if (tupleCollection != null) {
			throw new MedorException("A tuple collection has already been assigned to this PrefetchBuffer!!");
		}
		tupleCollection = tc;
	}

	public Tuple getTuple(Object index) throws MedorException {
		Map m = tupleMap;
		if (m == null) {
			return null;
		}
		Integer i = (Integer) m.get(index);
		if (i == null) {
			return null;
		}
		m.remove(index);
		int row = i.intValue();
		int oldrow = -1;
		Tuple res = null;
		//local copy to avoid the synchronization
		TupleCollection localTC = tupleCollection;
		if (localTC == null)
			return null;
		try {
			oldrow = localTC.getRow();
			res = localTC.getTuple(row);
		} catch(MedorException e) {
			//The tupleCollection has been closed or it is not possible to go
			// to the row specified
			res = null;
		} finally {
			if (multithread) {
				// try to restore old position into tupleCollection if the
				// tupleCollection is alive
				if (tupleCollection != null && oldrow != row) {
//TODO : Clone the res variable
					try {
						localTC.row(oldrow);
					} catch (MedorException e) {
						// ARRGH: cannot do anything more!!
					}
				}
			}
		}
		if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
			logger.log(BasicLevel.DEBUG, "Tuple found :"
					+ "\n\trow=" + row + "\toldrow=" + oldrow
					+ "\n\tindex=" + index
					+ "\n\ttuple=" + PrefetchCacheImpl.printTuple(res));
		}
		return res;
	}

	public synchronized void close() throws MedorException {
		closed = true;
		if (tupleCollection != null) {
			if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
				logger.log(BasicLevel.DEBUG, "Closing PrefetchBuffer: " + this);
			}
			try {
				//don't close the TupleCollection yet, but invalidate it.
				((PrefetchBufferHolder)tupleCollection)
						.invalidatePrefetchBuffer();
			} finally {
				tupleCollection = null;
			}
		}
	}

	public boolean isClosed() {
		return closed;
	}
}
