/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.filter.jorm.lib;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.naming.api.PNameCoder;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.expression.api.MalformedExpressionException;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.api.TypingException;
import org.objectweb.medor.eval.lib.TuplePNameGetter;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.lib.BasicVariableOperand;
import org.objectweb.medor.expression.lib.BasicOperator;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

/**
 * A CompositePName is an Operator representing the construction of a PName
 * from a NamingContext and several Fields.
 * <p>
 * The NamingContext is provided as a ParameterOperand. The name of the
 * ParameterOperand is defined to be:
 * <ol><li>The fully qualified class name for the PName of the class itself</li>
 * <li>The fully qualified class name, to which the attribute name is
 * concatenated for a reference PName</li>
 * </ol>
 * The Fields are provided as an array of FieldOperand.
 * It relies on a TuplePNameGetter
 * (@see org.objectweb.medor.query.jorm.lib.TuplePNameGetter) to construct the
 * PName.

 * @author	<A HREF="mailto:alia.mourad@rd.francetelecom.com><b>
 *			Mourad Alia
 *			</b></A>
 *		<A HREF="mailto:alexandre.lefebvre@rd.francetelecom.com><b>
 * 			Alexandre Lefebvre
 </b></A>
 */
public class CompositePName extends BasicOperator implements Operator {

	protected String[] cofns = null;
	protected TuplePNameGetter myGetter = null;


    /**
     * Constructs a SinglePName from a ParameterOperand representing the
     * naming context, and an array of FieldOperand. The FieldOperands should
     * be built on JormFields.
     * @param fs is the array of FieldOperands from which the value is taken.
     * @param po is the ParameterOperand representing the naming context.
     * @param t is the PType of the PName created by this operator.
     * @param compositefieldnames is the compositename fields. This array
     * is linked to parameter fs (FieldOperand[]).
     * @throws MedorException when the number of FieldOperand is not equal to
     * the size of the proj parameter.
     */
    public CompositePName(FieldOperand[] fs,
                          String[] compositefieldnames,
                          Operand po,
                          PType t) throws MedorException {
		super(t);
		expressions = new Expression[fs.length + 1];
		expressions[0] = po;
		System.arraycopy(fs, 0, expressions, 1, fs.length);
        cofns = compositefieldnames;
        Map name2Index = new HashMap(fs.length);
		boolean debug = logger != null && logger.isLoggable(BasicLevel.DEBUG);
		if (debug) {
			logger.log(BasicLevel.DEBUG, "Composite PName for " + fs.length + " fields");
		}
        for (int i=0; i<fs.length; i++) {
            QueryTreeField qtf = (QueryTreeField) fs[i].getField();
			if (debug) {
				logger.log(BasicLevel.DEBUG, "Field " + qtf.getName());
				Field[] tsf = qtf.getQueryTree().getTupleStructure().getFields();
				logger.log(BasicLevel.DEBUG, "Fields of qt:");
				for (int jj = 0; jj < tsf.length; jj++) {
					logger.log(BasicLevel.DEBUG, "Field name " + tsf[jj].getName());
				}
			}
            int rank = qtf.getQueryTree().getTupleStructure().getFieldRank(qtf);
            name2Index.put(compositefieldnames[i], new Integer(rank));
        }
        myGetter = new TuplePNameGetter(name2Index, po, logger);
    }


	/**
	 * It retrieves an array of the composite field name. The order of this
	 * array matches to the order of the FieldOperand array returned by the
	 * method getField()
	 */
	public String[] getCompositeFieldName() {
		return cofns;
	}

    public ParameterOperand getPNameManagerParameter() {
        return (expressions[0] instanceof ParameterOperand
			? (ParameterOperand) expressions[0]
			: null);
    }

	// IMPLEMENTATION OF THE Expression INTERFACE //
	//--------------------------------------------//

	public Object clone(Object clone, Map obj2clone) throws CloneNotSupportedException {
		clone = super.clone(clone, obj2clone);
		((CompositePName) clone).cofns = cofns;
		//clone the tuplePNameGetter because the field indexes can change
		// independantly in the futur
		if (myGetter != null) {
			((CompositePName) clone).myGetter = new TuplePNameGetter(myGetter);
		}
		return clone;
	}

	/**
	 * Gets the operand in wich the result will be put in
	 * @return an Operand Object
	 * @throws IllegalStateException if this expression is not
	 * compiled.
	 */
	public Operand getResult() throws IllegalStateException {
		if (verified) {
			return result;
		} else
			throw new IllegalStateException("Can't get result of an uncompiled expression");
	}

	/**
	 *
	 */
	public Operand evaluate(ParameterOperand[] pos, Object o)
		throws ExpressionException {
		//assign the value of the NamingContext
		expressions[0].evaluate(pos, o);
		//gets the naming context
		if (result == null)
			result = new BasicVariableOperand(type);
		PNameCoder pnc = (PNameCoder) ((Operand) expressions[0]).getObject();
		try {
			PName pn = pnc.decodeAbstract(myGetter, o);
			if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
				logger.log(BasicLevel.DEBUG, "PName built: " + pn);
			}
			result.setValue(pn);
		} catch (PException pe) {
			throw new ExpressionException("Problem for builing composite PName from tuple", pe);
		}
		return result;
	}

	/**
	 * Checks the semantic integrity of an expression.
	 * It checks that all types are compatible and prepare the expression to be
	 * evaluable. It also creates buffers where stores the result. Notes that
	 * when evaluating there is no creation of new objects. This method changes
	 * the state of this expression object, it will be evaluable and not
	 * modifiable.
	 * @throws org.objectweb.medor.expression.api.TypingException when incompatible types error
	 * occurs.
	 * @throws org.objectweb.medor.expression.api.MalformedExpressionException if syntax error
	 */
	public Operand compileExpression()
		throws TypingException, MalformedExpressionException {
		if (expressions != null && expressions.length > 1) {
			result = new BasicVariableOperand(type);
			verified = true;
		} else
		// Operands not yet assigned
			throw new MalformedExpressionException("null children value");
		return result;
	}

	public FieldOperand[] getFields() {
		FieldOperand[] res = new FieldOperand[expressions.length - 1];
		for(int i=0; i<res.length; i++) {
			res[i] = (FieldOperand) expressions[i + 1];
		}
		return res;
	}

	//Implementation of Operator interface
	/**
	 * There is no Java operator associated to a PName
	 */
	public String getOperatorString() {
		return "compositePName";
	}
}
