/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre, S. Chassande-Barrioz
 */

package org.objectweb.medor.filter.lib;

import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.UnevaluableExpressionException;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.MalformedExpressionException;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.TypingException;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.lib.BasicVariableOperand;
import org.objectweb.medor.expression.lib.BasicOperator;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.type.lib.PTypeSpaceMedor;

import java.util.List;

/**
 * The MemberOf class is an operator testing whether the value of a list of
 * Fields (the first n operands, which are supposed to be FieldOperands or
 * ParameterOperands) are
 * equal to one of the values of another list of Fields (the second n operands,
 * which are supposed to be FieldOperands).
 * <p>As an example, consider the left list (fopA, fopB) and the right list
 * (fopC, fopD). Consider that fopA and fopB are FieldOperands for fields A and
 * B of QueryTree qt1 and that fopC and fopD are FieldOperands for fields C and
 * D of QueryTree qt2.
 * <p>Consider the following data:
 * <table border="1">
 * <tr><td colspan="2"><b>qt1</b></td><td colspan="2"><b>qt2</b></td></tr>
 * <tr><td width="40"><b>a</b></td><td width="40"><b>b</b></td><td width="40"><b>c</b></td><td width="40"><b>d</b></td></tr>
 * <tr><td>a1</td><td>b1</td><td>a1</td><td>b1</td></tr>
 * <tr><td>a2</td><td>b2</td><td>a3</td><td>b3</td></tr>
 * <tr><td>a3</td><td>b3</td><td>a3</td><td>b3</td></tr>
 * <tr><td></td><td></td><td>a4</td><td>b4</td></tr>
 * </table>
 * <p>(a1,b1) and (a3,b3), values of (a,b), are members of the list of values
 * for (c,d), and appear in the result only once. Even (a3,b3) which appears
 * twice in (c,d) is only once in the result. This would not have been the case
 * with a join.
 * <p>Note that, in the case of PNames, we only allow comparison of two
 * PNames, that is, the number of operands is supposed to be two and not more.
 */
public class MemberOf extends BasicOperator implements Operator {

	/**
	 * Constructor for the operator.
	 * The operand array of this operators are the union both list parameter (
	 * left + right)
	 * @param left a list of FieldOperands to test.
	 * @param right the list of FieldOperands of which the left ones should be
	 * members.
	 * @throws org.objectweb.medor.expression.api.MalformedExpressionException
	 */
	public MemberOf(List left, List right) throws MalformedExpressionException {
		super(PTypeSpaceMedor.BOOLEAN);
		int size = left.size();
		//check that they are the same size
		if (right.size() != size)
			throw new MalformedExpressionException(
				"Left list has size " + size +
				" whereas right list has size " + right.size() +
				". They should be the same.");
		//check that no two Fields of left and right are built on the same
		//QueryTree
		expressions = new Expression[size * 2];
		for(int i=0; i<size; i++) {
			expressions[i] = (Expression) left.get(i);
			expressions[i + size] = (Expression) right.get(i);
			if (expressions[i] instanceof FieldOperand) {
				QueryTree qt1 = ((QueryTreeField) ((FieldOperand)expressions[i]).getField())
					.getQueryTree();
				QueryTree qt2 = ((QueryTreeField) ((FieldOperand)expressions[i + size]).getField())
						.getQueryTree();
				if (!(qt1 instanceof QueryLeaf) && (qt1 == qt2))
					throw new MalformedExpressionException(
						"The fields may not be built on the same QueryTree " + qt1);
			}
		}
	}

	//Implementation of the Expression interface
	public Operand evaluate(ParameterOperand[] pos, Object o)
		throws ExpressionException {
		//memory evaluation is not supported yet
		throw new UnevaluableExpressionException(
			"Unevaluable Expression: MemberOf currently not implemented");
	}

	public Operand compileExpression()
		throws TypingException, MalformedExpressionException {
		if (expressions != null) {
			result = new BasicVariableOperand(type);
			verified = true;
		}
		else
		// Operands not yet assigned
			throw new MalformedExpressionException("null children value");
		return result;
	}

	public Operand getResult() throws IllegalStateException {
		if (verified) {
			return result;
		}
		else
			throw new IllegalStateException("Can't get result of an uncompiled expression");
	}

	//Implementation of the Operator interface
	public String getOperatorString() {
		return Operator.MEMBEROF;
	}
}
