/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.filter.lib;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.MalformedExpressionException;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.lib.BasicBinaryOperator;
import org.objectweb.medor.expression.lib.BasicOperand;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.jorm.lib.PNameField;

import java.util.Map;

/**
 * A NavigatorOperator typically represents a path expression.
 * <p>
 * Such an expression starts from a FieldOperand for which the origin field
 * is a JormField of a PName type (ref).
 * @author	<A HREF="mailto:alia.mourad@rd.francetelecom.com><b>
 *			Mourad Alia
 *			</b></A>
 *		<A HREF="mailto:alexandre.lefebvre@rd.francetelecom.com><b>
 * 			Alexandre Lefebvre
 </b></A>
 */
public class NavigatorOperator extends BasicBinaryOperator {

	//TODO move this class to the filter.jorm.lib package
    protected boolean isPName;
    private MetaObject jormMO;


    /**
     * Creates a NavigatorOperator from a FieldOperand and an attribute name.
     * The origin field of this FieldOperand must be a JormField of a PName
     * type (or a PropagatedField pointing to such a JormField).
     * @param fop is the FieldOperand corresponding to the left operand.
     * @param att is the name of the attribute to which the operator nagivates.
     * @throws MedorException when fop does not reference a PNameField
     * or when no field att is reachable by the PNameField.
     */
    public NavigatorOperator(FieldOperand fop, String att)
		throws MedorException {
		super(fop, new BasicOperand(att));

        //Fetch the PNameField origin
        Field f = fop.getField();
        if (f instanceof PropagatedField) {
            Field[] fs = ((PropagatedField) f).getOriginFields();
            f = fs[0];
        }
        if (!(f instanceof PNameField))
            throw new MedorException(
                "Impossible to create a NavigatorOperator without PNameField "
                + "as begin of expression: " + f);
        PNameField pnf = (PNameField) f;
        //System.out.println("pnf=" + pnf);


        if (pnf.isClassPName()) {
            if (pnf.isInGenClass())
                throw new MedorException("Impossible to navigate with a GenClass: pnf=" + pnf.getName() + " /  attribut=" + att);
            //System.out.println("pnf.isClassPName");
            //The PName is the PName of the class
            jormMO = getTypedElement(pnf.getMetaObjectClass(), att);
            type = ((TypedElement) jormMO).getType();
        }
        else {
            //System.out.println("!pnf.isClassPName");
            //The PName is the PName of the reference field
            jormMO = pnf.getReference();
            if (jormMO instanceof ClassRef) {
                jormMO = getTypedElement(((ClassRef) jormMO).getMOClass(), att);
                //System.out.println("jormMO=" + jormMO);
                type = ((TypedElement) jormMO).getType();
            }
            else if (jormMO instanceof GenClassRef) {
                //Be carefull it is not possible to navigate after a gen class.
                type = ((GenClassRef) jormMO).getType();
            }
        }
        if (jormMO == null)
            throw new MedorException("No field " + att + " found");
        isPName = jormMO instanceof Reference;
    }

	public Object clone(Object clone, Map obj2clone) throws CloneNotSupportedException {
		clone = super.clone(clone, obj2clone);
		((NavigatorOperator) clone).isPName = isPName;
		((NavigatorOperator) clone).jormMO = jormMO;
		return clone;
	}

    /**
     * Creates a NavigatorOperator from a previously existing NavigatorOperator
     * and an attribute name.
     * @param nop is the NavigatorOperator corresponding to the left operand.
     * @param att is the name of the attribute to which the operator nagivates.
     * @throws MedorException when nop does not match to a PName
     * or when no field att is reachable by the PName.
     */
    public NavigatorOperator(NavigatorOperator nop, String att)
		throws MedorException {
        super(nop, new BasicOperand(att));
        if (!nop.isPName())
            throw new MedorException(
                "The specified Navigator operand is not a PName");

        jormMO = nop.getMetaObject();
        if (jormMO instanceof ClassRef) {
            jormMO = ((ClassRef) jormMO).getMOClass().getTypedElement(att);
            type = ((TypedElement) jormMO).getType();
        }
		//TODO allow nagivating GenClasses
        else if (jormMO instanceof GenClassRef) {
            throw new MedorException("Impossible to navigate with a GenClass: pnf=" + nop + " /  attribut=" + att);
        }
        if (jormMO == null)
            throw new MedorException("No field " + att + " found");
        isPName = jormMO instanceof Reference;
    }

    /**
     * Returns 'true' if the current path designes a field which is a
     * reference. It returns 'false' if the current path designes a primitive
     * field.
     */
    public boolean isPName() {
        return isPName;
    }

    /**
     * Returns the Jorm Meta object which describes the field that is
     * reached by this navigator. The jorm field is never null.
     */
    public MetaObject getMetaObject() {
        return jormMO;
    }

    /**
     * Returns the String representing the operator in Java
     * @return the String representing the operator in Java
     */
    public String getOperatorString() {
        return Operator.NAV;
    }

    /**
     * This operator is not evaluable, and must be transformed into JoinProject
     * instances during the optimization step.
     */
    public org.objectweb.medor.expression.api.Operand evaluate(ParameterOperand[] pos, Object o)
        throws ExpressionException {
        throw new ExpressionException("This operator is not evaluable, and must be "
            + "transformed into JoinProjects instance during the optimization step.");
    }

    /**
     * Checks the semantic integrity of an expression.
     * <p>It checks that all types are compatible and prepare the expression to
	 * be evaluable.It also creates buffers where stores the result. Notes that
     * when evaluating there is no creation of new objects. This method change
     * the state of this expression object, it will be evaluable and not
     * modifiable.
     * @throws org.objectweb.medor.expression.api.MalformedExpressionException if syntax error
     */
    public Operand compileExpression() throws MalformedExpressionException {
        throw new MalformedExpressionException("This operator is not evaluable,"
            + " and must be transform edinto JoinProjects instance during the"
            + " optimization step.");
    }

    private TypedElement getTypedElement(Class clazz, String fn) {
        int idx = fn.length();
        TypedElement te = null;
        te = clazz.getTypedElement(fn);
        while ((idx = fn.lastIndexOf(".", idx)) != -1 && te == null) {
            te = clazz.getTypedElement(fn.substring(idx + 1, fn.length()));
            idx--;
        }
        return te;
    }
}
