/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2004 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */
package org.objectweb.medor.optim.jorm;

import org.objectweb.jorm.mapper.rdb.metainfo.RdbClassMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbFilter;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbMapping;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ParentClassMapping;
import org.objectweb.jorm.naming.api.NamingFilterKeyProvider;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.lib.ExpressionPrinter;
import org.objectweb.medor.expression.lib.Or;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.jorm.api.JormExtent;
import org.objectweb.medor.query.jorm.lib.ClassExtent;
import org.objectweb.medor.query.lib.SelectProject;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * In the case of filtered inheritance, this rule replaces a QueryTree built on
 * top of JormExtents with inheritance by a QueryTree with JormExtents without
 * inheritance, on top of which are inserted new SelectProjects containing the
 * filter corresponding to the inheritance filter.
 * <p>MEDOR constructs its filter using the ones in RdbClassMapping, either
 * accessible through getRdbFilter(), if it has been provided by the user, or
 * computed from the NameDef filter otherwise (method to be provided in JORM).
 * </p>
 * @author A. Lefebvre
 */
public class JormFilteredInheritance extends JormRule {

	public JormFilteredInheritance() {
		super("JormFilteredInheritance");
	}

	public QueryTree rewrite(QueryTree qt, QueryNode parent)
			throws MedorException {
		if ((qt instanceof ClassExtent)
			&& ((ClassExtent) qt).getPMapper().getMapperName().startsWith("rdb")
			) {
			//obtain the expression corresponding to the filteredExtent
			Expression inhfilt = filterForExtent(
					(ClassExtent) qt,
					(Class) ((ClassExtent) qt).getMetaObject(),
					((ClassExtent) qt).getProjectName(),
					null,
					((JormExtent) qt).withSubClasses());
			if (debug) {
				log.log(BasicLevel.DEBUG, "Inheritance filter "
					+ ExpressionPrinter.e2str(inhfilt));
			}
			if (inhfilt != null) {
				//modify the Extent: do not consider inheritance any more
				((JormExtent) qt).setWithSubClasses(false);
				//create the new SelectProject with the filter
				QueryNode newSP = new SelectProject(inhfilt, "", null);
				Map localyReplaced = (parent == null ? null : new HashMap());
				//add corresponding propagated fields
				Field[] fs = qt.getTupleStructure().getFields();
				for (int i = 0; i < fs.length; i++) {
					Field newF = newSP.addPropagatedField(
						fs[i].getName(),
						fs[i].getType(),
						new QueryTreeField[]{(QueryTreeField) fs[i]});
					if (parent != null) {
						localyReplaced.put(fs[i], newF);
					}
				}
				//propagate changes to the parent
				replaceUsage(parent, localyReplaced);
			} // else (no filter found) then do nothing

		} else if (qt instanceof QueryNode) {
			QueryTree[] children = ((QueryNode) qt).getChildren();
			for (int i = 0; i < children.length; i++) {
				rewrite(children[i], (QueryNode) qt);
			}
		}
		return qt;
	}

	/**
	 * Calculates the filter expression corresponding to the filtered
	 * inheritance mapping.
	 * @param extent the extent
	 * @param currClass the current Class
	 * @param projectName the project name
	 * @param filter the filter currently computed
	 * @return the expression corresponding to the inheritance filter for that
	 * class
	 * @throws MedorException if concrete class without RdbFilter or if the
	 * NameDef differs for some of the subclasses.
	 */
	private Expression filterForExtent(
			ClassExtent extent,
			Class currClass,
			String projectName,
			Expression filter,
			boolean withSubClass)
			throws MedorException {
		//add the filter of the current class, if concrete class
		if (!currClass.isAbstract()) {
			if (debug) {
				log.log(BasicLevel.DEBUG, "Concrete class " + currClass.getName());
			}
			RdbMapping mapping = (RdbMapping) currClass
					.getClassProject(projectName)
					.getMapping("rdb");
			RdbClassMapping rdbCM = (RdbClassMapping) mapping.getClassMapping();

			//Fetch the mapping filter
			RdbFilter mapFilter = rdbCM.getRdbFilter();
			Expression localFilter = null;
			if (mapFilter != null) { //there is a Rdbfilter
				localFilter = mapFilter.getExpression();
				if (debug) {
					log.log(BasicLevel.DEBUG, "Mapping filter is "
						+ ExpressionPrinter.e2str(localFilter));
				}
			}
			if (localFilter == null) {
				//No mapping filter specified ==> check if the mode is
				// filtered mapping
				if (mapping.isFilteredMapping()) {
					Object pcm = extent.getPMapper().lookup(currClass.getFQName());
					try {
						if (pcm instanceof NamingFilterKeyProvider) {
							localFilter = rdbCM.getMappingFilterFromNamingFilter(
								(NamingFilterKeyProvider) pcm);
						} else {
							localFilter = rdbCM.getMappingFilterFromNamingFilter();
						}
					} catch (ExpressionException e) {
						throw new MedorException(e);
					}
					if (debug) {
						log.log(BasicLevel.DEBUG, "Naming filter is "
							+ ExpressionPrinter.e2str(localFilter));
					}
				}
				if (localFilter == null) {
					return null;
				}
			}
			//replace the ParameterOperand of the JORM filter with FieldOperands
			localFilter = modifiesJormFilterForExtent(localFilter, extent);

			//Merge the current filter with the one specified in parameter
			if (filter == null) {
				filter = localFilter;
			} else {
				filter = new Or(filter, localFilter);
			}
		}
		if (withSubClass) {
			//iterate over subclasses in order to have their filter too
			Iterator subclassesIt = currClass.getSubClasses().iterator();
			if (debug) {
				log.log(BasicLevel.DEBUG, "Subclasses for " + currClass.getName());
			}
			while (subclassesIt.hasNext()) {
				Class subcl = (Class) subclassesIt.next();
				if (debug) {
					log.log(BasicLevel.DEBUG, "Recursing for subclass "
						+ subcl.getName() + " of class " + currClass.getName());
				}
				filter = filterForExtent(
						extent,
						subcl,
						projectName,
						filter,
						true);
			}
		}
		return filter;
	}

	/**
	 * Takes the JORM filter as input, expressed with ParameterOperands, and
	 * replaces the ParameterOperands with FieldOperands on the ClassExtent.
	 * <p>If the field is not present on the Extent, add it.</p>
	 * @param filter The JORM filter expressed using ParameterOperands
	 * @param extent The ClassExtent
	 * @throws MedorException when adding the field to the Extent.
	 */
	private Expression modifiesJormFilterForExtent(Expression filter,
			ClassExtent extent)
			throws MedorException {
		if (filter instanceof Operator) {
			Operator op = (Operator) filter;
			int opnum = op.getOperandNumber();
			for (int i = 0; i < opnum; i++) {
				op.setExpression(i,
					modifiesJormFilterForExtent(op.getExpression(i), extent));
			}
			return op;
		} else if (filter instanceof ParameterOperand) {
			String fieldName = ((ParameterOperand) filter).getName();
			Field f = null;
			try {
				f = extent.getField(extent.getJormName() + "." + fieldName);
				if (debug) {
					log.log(BasicLevel.DEBUG, "Field already on extent: " + fieldName);
				}
			} catch (MedorException me) {
				if (debug) {
					log.log(BasicLevel.DEBUG, "Catched exception: " + me + " (field not on extent, will be added)");
				}
				f = extent.addField(((Class)extent.getMetaObject()).getTypedElement(fieldName));
				if (debug) {
					log.log(BasicLevel.DEBUG, "Added field to extent: " + fieldName + " " + f);
				}
			}
			return new BasicFieldOperand(f);
		} else {
			return filter;
		}
	}
}
