/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */

package org.objectweb.medor.optim.jorm;

import org.objectweb.medor.api.Field;
import org.objectweb.medor.expression.api.MalformedExpressionException;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.filter.jorm.lib.CompositePName;
import org.objectweb.medor.filter.jorm.lib.EncodePName;
import org.objectweb.medor.filter.jorm.lib.SinglePName;
import org.objectweb.medor.expression.lib.And;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.filter.lib.MemberOf;
import org.objectweb.medor.expression.lib.NotEqual;
import org.objectweb.medor.optim.lib.BasicRule;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.query.api.PropagFromNestedField;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.lib.Project;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * The aim of this rule is to remove comparaisons between PNames and move
 * up the Decode operators (i.e. SinglePName and CompositePName) at the root of
 * the QueryTree.
 * <p>
 * This rule is applied in one shot on the tree.
 * <p>The Decode operators (SinglePName and CompositePName) are produced by
 * the JormLeafRewriter rule.
 * @author S. Chassande-Barrioz
 * @see JormLeafRewriter
 */
public class JormGoUpDecodeRule extends BasicRule {

	public JormGoUpDecodeRule() {
		super("JormGoUpDecodeRule");
	}

	/**
	 * Rewrites a QueryTree by removing comparisons between PNames and moving
	 * up the Decode operators at the root of the QueryTree.
	 * <p>It relies on the goUpDecode method to perform the real work.
	 * @param qt the QueryTree to be rewritten
	 * @return the rewritten QueryTree
	 * @throws MedorException
	 */
	public QueryTree rewrite(QueryTree qt, QueryNode _parent)
			throws MedorException {
		debug = log != null && log.isLoggable(BasicLevel.DEBUG);
		Field[] fields = qt.getTupleStructure().getFields();
		if (!(qt instanceof QueryNode))
			return qt;
		//Performs the "real" work
		Map map = goUpDecode((QueryNode) qt, true);
		if (map.size() == 0)
			return qt;
		QueryTree result = new Project("", null);
		//On the root node, the useful decode operators are added.
		for (Iterator it = map.entrySet().iterator(); it.hasNext();) {
			Map.Entry me = (Map.Entry) it.next();
			Field decodefield = (Field) me.getKey();
			//debug("root: decodefield=" + decodefield.getName());
			Object neos = me.getValue();
			//Fetch the expression of the old decodefield
			Expression e = getCalculatedField(decodefield).getExpression();
			//Re-assign the field in the FieldOperand
			if (e instanceof CompositePName) {
				FieldOperand[] fos = ((CompositePName) e).getFields();
				String[] cofns = ((CompositePName) e).getCompositeFieldName();
				for (int i = 0; i < fos.length; i++) {
					fos[i].setField((Field) ((Map) neos).get(cofns[i]));
				}
			}
			else if (e instanceof SinglePName) {
				((SinglePName) e).getField().setField((Field) neos);
			}
			else
				throw new MedorException("Umanaged case: PName field decoder in complex expression");
			//Add the field on the root
			((QueryNode) result).addCalculatedField(
				decodefield.getName(),
				decodefield.getType(),
				e);
		}
		//Propagate the other fields
		for (int i = 0; i < fields.length; i++) {
			String fn = fields[i].getName();
			if (qt.getTupleStructure().contains(fn)
				&& !result.getTupleStructure().contains(fn)) {
				QueryTreeField prev = (QueryTreeField)
					qt.getTupleStructure().getField(fn);
				debug("root: propag=" + prev.getName());
				((QueryNode) result).addPropagatedField(
					prev.getName(),
					prev.getType(),
					new QueryTreeField[]{prev});
			}
		}
		return result;
	}

	/**
	 * This recursive method performs the real work of the rule.
	 * @param qn is the QueryNode on which the rule must be applied.
	 * @param isRoot is a boolean which indicates if the qn is the root of the
	 * QueryTree.
	 * @return a map with the following characteristics:
	 * <ul><li>
	 * key = an old projected field of a child. The old field is a PName decode
	 * or a Propagated field.</li>
	 * <li>value = A map of composite field name associated to the new projected
	 * fields of the child. These fields replace the PName identifier. They are
	 * the PName fields. (key=composite fname, value=field)
	 * </li></ul>
	 */
	public Map goUpDecode(QueryNode qn, boolean isRoot) throws MedorException {
		Map childMap = new HashMap();
		Map map = new HashMap();
		//Call children and collaspe the result into a single map
		QueryTree[] qts = qn.getChildren();
		for (int i = 0; i < qts.length; i++) {
			if (qts[i] instanceof QueryNode)
				childMap.putAll(goUpDecode((QueryNode) qts[i], false));
		}

		//Use the map of the children to replace in filter the FieldOperands
		//which reference (directly or not) a PName field.
		Expression filter = qn.getQueryFilter();
		if (filter != null) {
			debug("replaceFieldOperand in filter");
			try {
				qn.setQueryFilter(replaceFieldOperand(filter, childMap).e);
			}
			catch (ExpressionException e) {
				throw new MedorException(e);
			}
		}

		// Replace the fields which used the pname and fill a new map with the
		// couples:
		// (old_field_in_this_node, Map(composite field name, new_fields_in_this_node))
		Field[] fields = qn.getTupleStructure().getFields();
		for (int i = 0; i < fields.length; i++) {
			debug("Field " + fields[i].getName() + " " + fields[i].getType());
			// replace the decode into CalculatedFields
			if (!isRoot && fields[i] instanceof CalculatedField) {
				Expression e = ((CalculatedField) fields[i]).getExpression();
				if (e instanceof CompositePName) {
					debug("CalculatedField(CompositePName): begin");
					qn.removeField(fields[i].getName());
/*
					if (qn.getTupleStructure().contains(fields[i].getName())) {
						throw new MedorException("Field not correctly removed.");
					}
*/
					FieldOperand[] fops = ((CompositePName) e).getFields();
					String[] cofns =
						((CompositePName) e).getCompositeFieldName();
					Map al = new HashMap(fops.length);
					debug("CompositePName: " + fops.length);
					for (int j = 0; j < fops.length; j++) {
						Field prev = fops[j].getField();
						Field neo = null;
						if (!qn.getTupleStructure().contains(prev.getName())) {
							debug("Field " + prev.getName() + " of CompositePName of not already projected.");
							neo = qn.addPropagatedField(
								prev.getName(),
								prev.getType(),
								new QueryTreeField[]{(QueryTreeField) prev});
						}
						else {
							debug("Field " + prev.getName() + " of CompositePName was already propagated.");
							neo =
								qn.getTupleStructure().getField(prev.getName());
						}
						debug("put(" + cofns[j] + "," + neo + ")");
						al.put(cofns[j], neo);
					}
					// Put in the result the association between the removed
					// calculated field and the new fields
					map.put(fields[i], al);
					debug("CalculatedField(CompositePName): end");
				}
				else if (e instanceof SinglePName) {
					debug("CalculatedField(SinglePName): begin");
					qn.removeField(fields[i].getName());
					Field prev = ((SinglePName) e).getField().getField();
					Field neo = null;
					if (!qn.getTupleStructure().contains(prev.getName()))
						neo = qn.addPropagatedField(
							prev.getName(),
							prev.getType(),
							new QueryTreeField[]{(QueryTreeField) prev});
					else
						neo = qn.getTupleStructure().getField(prev.getName());
					// Put in the result the association between the removed
					// calculated field and the new fields
					map.put(fields[i], neo);
					debug("CalculatedField(SinglePName): end");
				}
			}

			//replace the propagated fields which project a PName.
			else if (fields[i] instanceof PropagatedField) {
				debug("PropagatedField: begin");
				Field[] prev =
					((PropagatedField) fields[i]).getPreviousFields();
				for (int j = 0; j < prev.length; j++) {
					Object o = childMap.get(prev[j]);
					//debug("prev[j]= " + prev[j].getName() + " / al=" + al);
					//Is the field uses a removed field of children
					if (o != null) {
						//replacing ...
						qn.removeField(fields[i].getName());
						if (o instanceof Map) {
							Map neos = new HashMap();
							for (Iterator it = ((Map) o).entrySet().iterator();
								 it.hasNext();) {
								Map.Entry me = (Map.Entry) it.next();
								Field p = (Field) me.getValue();
								Field neo = null;
								if (!qn.getTupleStructure().contains(
									p.getName()))
									neo = qn.addPropagatedField(
										p.getName(),
										p.getType(),
										new QueryTreeField[]
										{(QueryTreeField) p});
								else
									neo = qn.getTupleStructure().getField(
										p.getName());
								neos.put(me.getKey(), neo);
							}
							// Put in the result the association between the
							//removed spropagated field and the new fields
							map.put(fields[i], neos);
						}
						else if (o instanceof Field) {
							Field p = (Field) o;
							Field neo = null;
							if (!qn.getTupleStructure().contains(p.getName()))
								neo = qn.addPropagatedField(
									p.getName(),
									p.getType(),
									new QueryTreeField[]{(QueryTreeField) p});
							else
								neo = qn.getTupleStructure().getField(
									p.getName());
							map.put(fields[i], neo);
						}
					}
				}
				debug("PropagatedField: end");
			}
		}
		return map;
	}

	/**
	 * Replaces PName FieldOperands in an Expression with Operands on the
	 * the corresponding Fields
	 * @param e the Expression in which to replace the PName FieldOperands
	 * @param map the Map, produced by goUpDecode, between the PNames and the
	 * Fields composing the PNames (see goUpDecode)
	 * @return a ModifiedExpression
	 * @throws MedorException if PNames are incompatible
	 */
	private ModifiedExpression replaceFieldOperand(Expression e, Map map)
		throws MedorException, ExpressionException {
		//debug("replaceFieldOperand: ", e, map);
		ModifiedExpression me = new ModifiedExpression();
		if (e instanceof Equal
			|| e instanceof NotEqual
			|| e instanceof MemberOf) {
			Operator bo = (Operator) e;
			Expression le = bo.getExpression(0);
			Expression re = bo.getExpression(1);
			Object lf = null;
			if (le instanceof FieldOperand) {
				lf = map.get(((FieldOperand) le).getField());
			}
			Object rf = null;
			if (re instanceof FieldOperand) {
				rf = map.get(((FieldOperand) re).getField());
			}
			if (lf != null && rf != null) {
				if (!lf.getClass().equals(rf.getClass()))
					throw new MedorException("incompatible PNames (single versus composite)");
				Expression expr = null;
				if (lf instanceof Map) {
					debug("lf=" + lf);
					//Composite name
					if (e instanceof MemberOf) {
						ArrayList l1 =
							new ArrayList(((Map) lf).entrySet().size());
						ArrayList l2 =
							new ArrayList(((Map) lf).entrySet().size());
						for (Iterator it = ((Map) lf).entrySet().iterator();
							 it.hasNext();) {
							Map.Entry mape = (Map.Entry) it.next();
							l1.add(mape.getValue());
							l2.add(((Map) rf).get(mape.getKey()));
						}
						if (me.e == null)
							me.e = new MemberOf(l1, l2);
						else
							me.e = new And(new MemberOf(l1, l2), me.e);
					}
					else
						for (Iterator it = ((Map) lf).entrySet().iterator();
							 it.hasNext();) {
							Map.Entry mape = (Map.Entry) it.next();
							expr = getExpr(e,
								new BasicFieldOperand((Field) mape.getValue()),
								new BasicFieldOperand((Field)
								((Map) rf).get(mape.getKey()))
							);
							if (me.e == null)
								me.e = expr;
							else
								me.e = new And(expr, me.e);
						}
				}
				else {
					//single name
					expr = getExpr(e,
						new BasicFieldOperand((Field) lf),
						new BasicFieldOperand((Field) rf));
					if (me.e == null)
						me.e = expr;
					else
						me.e = new And(expr, me.e);
				}
				me.isModified = true;
				return me;
			}
			else if (le instanceof ParameterOperand && rf != null) {
				Expression expr = null;
				if (rf instanceof Map) {
					for (Iterator it = ((Map) rf).entrySet().iterator();
						 it.hasNext();) {
						Map.Entry mape = (Map.Entry) it.next();
						Field f = (Field) mape.getValue();
						expr = getExpr(e,
							new EncodePName(
								f.getType(),
								((ParameterOperand) le).getName(),
								(String) mape.getKey()),
							new BasicFieldOperand(f)
						);
						if (me.e == null)
							me.e = expr;
						else
							me.e = new And(expr, me.e);
					}
				}
				else {
					Field f = (Field) rf;
					expr = getExpr(e,
						new EncodePName(
							f.getType(),
							((ParameterOperand) le).getName()),
						new BasicFieldOperand(f)
					);
					if (me.e == null)
						me.e = expr;
					else
						me.e = new And(expr, me.e);
				}
				me.isModified = true;
				return me;
			}
			else if (re instanceof ParameterOperand && lf != null) {
				Expression expr = null;
				if (lf instanceof Map) {
					for (Iterator it = ((Map) lf).entrySet().iterator();
						 it.hasNext();) {
						Map.Entry mape = (Map.Entry) it.next();
						Field f = (Field) mape.getValue();
						expr = getExpr(e,
							new BasicFieldOperand(f),
							new EncodePName(
								f.getType(),
								((ParameterOperand) re).getName(),
								(String) mape.getKey())
						);
						if (me.e == null)
							me.e = expr;
						else
							me.e = new And(expr, me.e);
					}
				}
				else {
					Field f = (Field) lf;
					expr = getExpr(e,
						new BasicFieldOperand(f),
						new EncodePName(
							f.getType(),
							((ParameterOperand) re).getName())
					);
					if (me.e == null)
						me.e = expr;
					else
						me.e = new And(expr, me.e);
				}
				me.isModified = true;
				return me;
			}
		}
		if (e instanceof Operator) {
			Operator op = (Operator) e;
			me = new ModifiedExpression(e, false);
			ModifiedExpression _me = null;
			for (int i = 0; i < op.getOperandNumber(); i++) {
				_me = replaceFieldOperand(op.getExpression(i), map);
				if (_me.isModified) {
					me.isModified = true;
					op.setExpression(i, _me.e);
				}
			}
		}
		else {
			me.isModified = false;
		}
		me.e = e;
		return me;
	}

	/**
	 *
	 * @param operator
	 * @param loperand
	 * @param roperand
	 * @return
	 */
	private Expression getExpr(Expression operator,
							   Expression loperand,
							   Expression roperand)
		throws MalformedExpressionException {
		if (operator instanceof Equal)
			return new Equal(loperand, roperand);
		else if (operator instanceof NotEqual)
			return new NotEqual(loperand, roperand);
		else {//it is a MemberOf
			ArrayList l1 = new ArrayList(1);
			l1.add(loperand);
			ArrayList l2 = new ArrayList(1);
			l2.add(roperand);
			return new MemberOf(l1, l2);
		}
	}

	/**
	 * It tries to reach the CalculatedField with a field.
	 * @param f
	 * @throws MedorException
	 */
	private CalculatedField getCalculatedField(Field f) throws MedorException {
		if (f == null)
			return null;
		if (f instanceof PropagatedField)
			return getCalculatedField(
				((PropagatedField) f).getPreviousFields()[0]);
		else if (f instanceof CalculatedField)
			return (CalculatedField) f;
		else if (f instanceof PropagFromNestedField)
			return getCalculatedField(
				((PropagFromNestedField) f).getPreviousFlatField()[0]);
		else
			throw new MedorException("Medor Internal error: Unknown field");
	}
    private void debug(String msg) {
        //log.log(BasicLevel.DEBUG, msg);
    }

/*
    private static void debug(String msg, Expression e, Map map) {

        //msg += " / expression=" + TestExpressionHelper.e2str(e);
        msg += " / map(";
        Iterator it = map.entrySet().iterator();
        while (it.hasNext()) {
            Map.Entry me = (Map.Entry) it.next();
            msg += "(" + ((Field) me.getKey()).getName() + "=>" + me.getValue() + ")";
        }
        msg += ")";
        debug(msg);

    }
*/
}
