/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */

package org.objectweb.medor.optim.jorm;

import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.lib.JormPathHelper;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassMapping;
import org.objectweb.jorm.metainfo.api.ParentClassMapping;
import org.objectweb.jorm.metainfo.api.ClassProject;
import org.objectweb.jorm.metainfo.api.GenClassMapping;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.Mapping;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.PrimitiveElementMapping;
import org.objectweb.jorm.naming.api.PNameCoder;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.lib.BasicOperand;
import org.objectweb.medor.optim.lib.BasicRule;
import org.objectweb.medor.query.jorm.api.JormExtent;
import org.objectweb.medor.query.jorm.lib.ClassExtent;
import org.objectweb.medor.query.jorm.lib.GenClassExtent;
import org.objectweb.medor.query.jorm.lib.PNameField;
import org.objectweb.medor.type.lib.PTypeSpaceMedor;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * This class is a tool for the management of the JormExtent.
 * <p>It is used for the
 * replacement of JormExtent instances into specific QueryLeaf instances.
 *
 * @author S.Chassande-Barrioz
 */
public abstract class JormRule extends BasicRule {

    public JormRule(String suffix) {
        super(suffix);
    }


    /**
     * Retrieves the primitive element corresponding to a field name of
     * a jorm object (class or generic class).
     * @param extent represents the Jorm persistent class and references the
     * Jorm meta object containing the expected jorm meta object.
     * @param fieldname is the name of the
     * @throws MedorException if the field does not exist.
     */
    protected PrimitiveElement getPrimitiveElement(JormExtent extent,
                                                   String fieldname)
        throws MedorException {
        PrimitiveElement te = null;
        if (extent.getMetaObject() instanceof Class) {
            Class clazz = (Class) extent.getMetaObject();
            te = (PrimitiveElement) clazz.getTypedElement(fieldname);
            if (te == null) {
                te = clazz.getHiddenField(fieldname);
            }
            if (te == null) {
                String msg = "No field " + fieldname
                    + " available on the class " + clazz.getFQName();
                log.log(BasicLevel.ERROR, msg);
                throw new MedorException(msg);
            }
        }
        else if (extent.getMetaObject() instanceof GenClassRef) {
            GenClassRef gcr = (GenClassRef) extent.getMetaObject();
            te = gcr.getHiddenField(fieldname);
            if (te == null) {
                String msg = "No field " + fieldname
                    + " available on the generic class "
                    + gcr.getGenClassId();
                log.log(BasicLevel.ERROR, msg);
                throw new MedorException(msg);
            }
        }
        else {
            throw new MedorException("unamanged jorm extent: " + extent);
        }
        return te;
    }

    /**
     * Retrieves the RdbPrimitiveElementMapping of a field
     * @param extent the extent of a persistent class (ClassExtent
     * or GenClassExtent)
     * @param fieldName is the name of the field which the mapping is expected
     * @throws MedorException if the extent is not managed or if the field does
	 * not have a mapping.
     */
    protected PrimitiveElementMapping getPEM(JormExtent extent,
                                              String fieldName)
        throws MedorException {
        PrimitiveElementMapping res = null;
        String pn = extent.getProjectName();
        String mn = extent.getPMapper().getMapperName();
        if (extent instanceof ClassExtent) {
            Class clazz = (Class) extent.getMetaObject();
            ClassMapping cm = getClassMapping(extent);
			String jormFieldName = getJORMFieldName(fieldName, extent);
            res = cm.getPrimitiveElementMapping(jormFieldName);
			if (res == null) {
				ArrayList pcms = new ArrayList(cm.getParentClassMappings());
				while (!pcms.isEmpty() && res == null) {
					ParentClassMapping pcm = (ParentClassMapping) pcms.remove(0);
					ClassMapping _cm = pcm.getMOClass()
						.getClassProject(pcm.getProjectName())
						.getMapping(pcm.getMapperName()).getClassMapping();
					res = _cm.getPrimitiveElementMapping(jormFieldName);
					if (res == null) {
						pcms.addAll(_cm.getParentClassMappings());
					}
				}
			}
			if (res == null) {
                String msg = "No mapping for field '" + getJORMFieldName(fieldName, extent)
                    + "' of the class '" + clazz.getFQName()
                    + "' in the project '" + pn
                    + "' and the mapping '" + mn;
                log.log(BasicLevel.ERROR, msg);
                log.log(BasicLevel.ERROR, "Existing field mappings:");
                for(Iterator it= cm.getAllPrimitiveElementMappings().iterator(); it.hasNext();) {
                    res = (PrimitiveElementMapping) it.next();
                    log.log(BasicLevel.ERROR, "Field: "
                        + ((PrimitiveElement) res.getLinkedMO()).getName());
                }
                throw new MedorException(msg);
            }
            return res;
        }
        else if (extent instanceof GenClassExtent) {
            GenClassRef gcr = (GenClassRef) extent.getMetaObject();
            GenClassMapping gcm = getGenClassMapping(extent);
            res = gcm.getPrimitiveElementMapping(
                    getJORMFieldName(fieldName, extent));
            if (res == null) {
                String msg = "No mapping for field '" + getJORMFieldName(fieldName, extent)
                    + "' of the generic class '" + gcr.getGenClassId()
                    + "' in the project '" + pn
                    + " and the mapping '" + mn;
                log.log(BasicLevel.ERROR, msg);
                log.log(BasicLevel.ERROR, "Existing field mappings:");
                for(Iterator it= gcm.getPrimitiveElementMappings().iterator(); it.hasNext();) {
                    res = (PrimitiveElementMapping) it.next();
                    log.log(BasicLevel.ERROR, "Field: "
                        + ((PrimitiveElement) res.getLinkedMO()).getName());
                }
                throw new MedorException(msg);
            }
            return res;
        }
        else {
            throw new MedorException("Unmanaged JormExtent: " + extent);
        }
    }

    /**
     * Returns the field name without the extent name
     * @param fullName the full field name containing the extent name
     * @param ext the extent
      * @return the abridged field name
     */
    protected String getJORMFieldName(String fullName, JormExtent ext) {
        int idx = fullName.indexOf(ext.getName() + '.');
        if (idx == -1)
            return fullName;
        else
            return fullName.substring(ext.getName().length() + 1);
    }

    /**
     * Retrieves the jorm meta object describing the mapping of the persistent
     * object representing by its extent. This method must be call with a
     * ClassExtent instance. It retrieves a RdbClassMultiMapping instance.
     * The JormExtent instance must contain the project name and the mapper
     * instance.
     * @param extent must be a ClassExtent instance.
     * @return a RdbClassMultiMapping instance
     * @throws org.objectweb.medor.api.MedorException if no class mapping has been found for the project
     * name and the mapper specified by the JormExtent.
     */
    protected ClassMapping getClassMapping(JormExtent extent)
        throws MedorException {
        String pn = extent.getProjectName();
        PMapper mapper = extent.getPMapper();
        if (mapper == null) {
            String msg = "No mapper has not been assigned to the JormExtent (jorm name: "
                + extent.getJormName()
                + ", node name: " + extent.getName() + ")";
            log.log(BasicLevel.ERROR, msg);
            throw new MedorException(msg);
        }
        String mn = mapper.getMapperName();
        Class clazz = (Class) extent.getMetaObject();
        ClassProject cp = clazz.getClassProject(pn);
        if (cp == null) {
            String msg = "No project " + pn + " for the class "
                + clazz.getFQName();
            log.log(BasicLevel.ERROR, msg);
            throw new MedorException(msg);
        }
        int idx =  mn.indexOf('.');
        if (idx != -1) {
            mn = mn.substring(0, idx);
        }
        Mapping m = cp.getMapping(mn);
        if (m == null) {
            String msg = "No mapping '" + mn
                + "' in the project '" + pn
                + "' of the class " + clazz.getFQName();
            log.log(BasicLevel.ERROR, msg);
            throw new MedorException(msg);
        }
        return m.getClassMapping();
    }

    /**
     * Retrieves the jorm meta object describing the mapping of the persistent
     * object representing by its extent. This method must be call with a
     * GenClassExtent instance. It retrieves a RdbGenClassMapping instance.
     * The JormExtent instance must contain the project name and the mapper
     * instance.
     * @param extent must be a ClassExtent instance.
     * @return a RdbGenClassMapping instance
     * @throws org.objectweb.medor.api.MedorException if no class mapping has been found for the project
     * name and the mapper specified by the JormExtent.
     */
    protected GenClassMapping getGenClassMapping(JormExtent extent)
        throws MedorException {
        String pn = extent.getProjectName();
        String mn = extent.getPMapper().getMapperName();
        GenClassRef gcr = (GenClassRef) extent.getMetaObject();
        MetaObject parentclazz = gcr.getParent();
        while (!(parentclazz instanceof Class)) {
            parentclazz = parentclazz.getParent();
        }
        ClassProject cp = ((Class) parentclazz).getClassProject(pn);
        if (cp == null) {
            String msg = "No project " + pn + " for the generic class "
                + gcr.getGenClassId();
            log.log(BasicLevel.ERROR, msg);
            throw new MedorException(msg);
        }
        int idx =  mn.indexOf('.');
        if (idx != -1) {
            mn = mn.substring(0, idx);
        }
        Mapping m = cp.getMapping(mn);
        if (m == null) {
            String msg = "No mapping '" + mn
                + "' in the project '" + pn
                + "' of the generic class " + gcr.getGenClassId();
            log.log(BasicLevel.ERROR, msg);
            throw new MedorException(msg);
        }
        return m.getGenClassMapping(gcr.getGenClassId());
    }

    /**
     * Retrieves the Operand representing the PNameCoder needed to decode a
     * pname. If the mapper is assigned to the extent and the persistent class
     * is initialized then a constant operand is returned. Otherwise a
     * ParameterOperand is returned.
     * @param extent can contain the mapper
     * @param pnf is the field representing the PName to be decoded
     * @return a Operand instance.
     */
    protected Operand getPNCOperand(JormExtent extent, PNameField pnf) {
        PNameCoder pnc =  JormPathHelper.getPNameCoder(
            pnf.getPNamingContextParameter(), extent.getPMapper());
		log.log(BasicLevel.DEBUG, "getPNCOperand for pnf " + pnf);

        if (pnc == null) {
            return new org.objectweb.medor.expression.lib.BasicParameterOperand(
                PTypeSpaceMedor.NAMING_CONTEXT,
                pnf.getPNamingContextParameter());
        } else {
            return new BasicOperand(pnc, PTypeSpaceMedor.NAMING_CONTEXT);
        }
    }
}
