/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre, S. Chassande-Barrioz
 */
package org.objectweb.medor.optim.lib;

import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.lib.ExpressionPrinter;
import org.objectweb.medor.lib.Log;
import org.objectweb.medor.optim.api.RewriteRule;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.query.api.NestQueryNode;
import org.objectweb.medor.query.api.NestedField;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.util.ArrayList;
import java.util.Map;

/**
 * @author S. Chassande-Barrioz
 */
public abstract class BasicRule implements RewriteRule {

	/**
	 * A ModifiedExpression is a structure containing an Expression and a
	 * boolean indicating whether the Expression has been modified or not.
	 * <p>
	 * It is used by RewriteRules to check whether an Expression has been
	 * modified, in order to propagate the changes.
	 */
	public class ModifiedExpression {
		public boolean isModified;
		public Expression e;

		public ModifiedExpression() {
		}

		public ModifiedExpression(Expression ex, boolean ism) {
			e = ex;
			isModified = ism;
		}
	}

	protected Logger log = null;

	protected boolean debug = false;

	public BasicRule() {
		log = Log.loggerFactory.getLogger(
				Log.MEDOR_PREFIX + ".optim.rule");
		debug = log != null && log.isLoggable(BasicLevel.DEBUG);
	}

	public BasicRule(String suffix) {
		log = Log.loggerFactory.getLogger(
				Log.MEDOR_PREFIX + ".optim.rule." + suffix);
		debug = log != null && log.isLoggable(BasicLevel.DEBUG);
	}

	public QueryTree rewrite(QueryTree qt) throws MedorException {
		return rewrite(qt, null);
	}

	private boolean replace(Expression e, Map old2neo) {
		log.log(BasicLevel.DEBUG, "ReplaceUsage: Replacing expression " + ExpressionPrinter.e2str(e));
		if (e instanceof Operator) {
			int size = ((Operator) e).getOperandNumber();
			boolean res = false;
			for (int i = 0; i < size; i++) {
				res |= replace(((Operator) e).getExpression(i), old2neo);
			}
			return res;
		} else if (e instanceof FieldOperand) {
			FieldOperand fo = (FieldOperand) e;
			Field f = (Field) old2neo.get(fo.getField());
			if (f != null) {
				log.log(BasicLevel.DEBUG, "ReplaceUsage: Found new Field " + f);
				fo.setField(f);
				return true;
			}
		}
		return false;
	}

	protected void replaceUsage(QueryNode qn, Map old2neo)
			throws MedorException {
		if (qn == null) {
			return;
		}
		Field[] fs = qn.getTupleStructure().getFields();
		for (int i = 0; i < fs.length; i++) {
			if (fs[i] instanceof PropagatedField) {
				PropagatedField pf = (PropagatedField) fs[i];
				QueryTreeField parent = (QueryTreeField)
						pf.getPreviousFields()[0];
				log.log(BasicLevel.DEBUG, "ReplaceUsage: Replacing PropagatedField " + pf + " (" + parent + ")");
				QueryTreeField neo = (QueryTreeField) old2neo.get(parent);
				if (neo != null) {
					log.log(BasicLevel.DEBUG, "ReplaceUsage: with " + neo);
					qn.updatePropagatedField(fs[i].getName(),
							new QueryTreeField[]{neo});
				}
			} else if (fs[i] instanceof CalculatedField) {
				CalculatedField cf = (CalculatedField) fs[i];
				if (replace(cf.getExpression(), old2neo)) {
					qn.updateCalculatedField(
							cf.getName(), cf.getExpression());
				}
			} else if (fs[i] instanceof NestedField) {
				NestedField nf = (NestedField) fs[i];
				ArrayList newFields = new ArrayList();
				boolean changedNested = false;
				for (int j = 0; j < nf.getSize(); j++) {
					log.log(BasicLevel.DEBUG, "ReplaceUsage: Replacing NestedField field " + nf.getField(j+1) + " " + nf.getField(j+1).getName());
					QueryTreeField neo =
							(QueryTreeField) old2neo.get(nf.getField(j+1));
					if (neo != null) {
						log.log(BasicLevel.DEBUG, "Found as " + neo + " " + neo.getName() + " of node " + neo.getQueryTree());
						changedNested = true;
						if (! newFields.contains(neo)) {
							newFields.add(neo);
						}
					}
					else {
						log.log(BasicLevel.DEBUG, "Keeping old field");
						if (! newFields.contains(nf.getField(j+1))) {
							newFields.add(nf.getField(j+1));
						}
					}
				}
				if (changedNested) {
					log.log(BasicLevel.DEBUG, "Something changed");
					((NestQueryNode)qn).replaceNestedField(fs[i].getName(), (Field[]) newFields.toArray(new Field[] {}));
					}
			} else throw new MedorException("Unmanaged field: " + fs[i].getName());
		}
	}
}
