/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */
package org.objectweb.medor.optim.lib;

import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.api.BinaryLogicalOperator;
import org.objectweb.medor.expression.api.BinaryOperator;
import org.objectweb.medor.expression.api.Comparator;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.lib.And;
import org.objectweb.medor.expression.lib.ConditionalAnd;
import org.objectweb.medor.expression.lib.ConditionalOr;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.expression.lib.Greater;
import org.objectweb.medor.expression.lib.GreaterEqual;
import org.objectweb.medor.expression.lib.Lower;
import org.objectweb.medor.expression.lib.LowerEqual;
import org.objectweb.medor.expression.lib.Not;
import org.objectweb.medor.expression.lib.NotEqual;
import org.objectweb.medor.expression.lib.Or;
import org.objectweb.medor.filter.lib.IsNull;
import org.objectweb.medor.query.api.FilteredQueryTree;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class is a RewriteRule implementation which vists a tree of QueryTree
 * in order to push the not operator in the filters.
 * <p>
 * Somes examples:
 * <ul>
 * <li> !(a && b) is replaced by (!a || !b)</li>
 * <li>!(a > b) is replaced by (a <= b)</li>
 * <li>!(a = b) is replaced by (a != b)</li>
 * </ul>
 *
 * @author S. Chassande-Barrioz
 */

public class PushNotInExpressionRule extends BasicRule {

	public PushNotInExpressionRule() {
		super("PushNotInExpressionRule");
	}

	/**
	 * It modifies the expressions used as filter in a (tree of) QueryTree by pushing the
	 * not operator.
	 * Somes examples:
	 * !(a && b) is replaced by (!a || !b)
	 * !(a > b) is replaced by (a <= b)
	 * !(a = b) is replaced by (a != b)
	 * ...
	 * @param qt is the query on which the rule must be applied
	 * @throws MedorException if filters are malforrmed
	 */
	public QueryTree rewrite(QueryTree qt, QueryNode _parent)
			throws MedorException {
		if (qt instanceof FilteredQueryTree) {
			ModifiedExpression me = pushNotInExpression(
				((FilteredQueryTree) qt).getQueryFilter(), false);
			if (me.isModified)
				((FilteredQueryTree) qt).setQueryFilter(me.e);
		}
		if (qt instanceof QueryNode) {
			QueryTree[] children = ((QueryNode) qt).getChildren();
			for (int i = 0; i < children.length; i++) {
				rewrite(children[i], (QueryNode) qt);
			}
		}
		return qt;
	}

	/**
	 * This recursive method push the Not operator in an expression.
	 * @param e the current expression inside which the not operator must be
	 * pushed
	 * @param hasANot is boolean which indicates if a not is currently being
	 * push
	 * @throws MedorException if the expression is mal formed.
	 */
	public ModifiedExpression pushNotInExpression(
		Expression e, boolean hasANot) throws MedorException {
        log.log(BasicLevel.DEBUG, "pushNotInExpression: (" + hasANot + ", exp:" + e + ")");
		ModifiedExpression me = new ModifiedExpression(e, false);
		if (hasANot && e instanceof Comparator) {
			Expression l = ((BinaryOperator) e).getExpression(0);
			Expression r = ((BinaryOperator) e).getExpression(1);
			me.isModified = true;
			if (e instanceof Equal)
				me.e = new NotEqual(l, r);
			else if (e instanceof NotEqual)
				me.e = new Equal(l, r);
			else if (e instanceof Lower)
				me.e = new GreaterEqual(l, r);
			else if (e instanceof GreaterEqual)
				me.e = new Lower(l, r);
			else if (e instanceof LowerEqual)
				me.e = new Greater(l, r);
			else if (e instanceof Greater)
				me.e = new LowerEqual(l, r);
			else
				throw new MedorException(
					"Comparator unmanaged by this method: " + e);
		}
		else if (e instanceof Not)
			return pushNotInExpression(((Not) e).getExpression(), !hasANot);
		else if (e instanceof IsNull) {
            if (hasANot) {
                ((IsNull) me.e).setNot(!((IsNull) me.e).getNot());
                me.isModified = true;
            }
        } else if (e instanceof BinaryLogicalOperator) {
			ModifiedExpression lme = pushNotInExpression(
				((BinaryOperator) e).getExpression(0), hasANot);
			ModifiedExpression rme = pushNotInExpression(
				((BinaryOperator) e).getExpression(1), hasANot);
			if (hasANot) {
				me.isModified = true;
				if (e instanceof And)
					me.e = new Or(lme.e, rme.e);
				else if (e instanceof ConditionalAnd)
					me.e = new ConditionalOr(lme.e, rme.e);
				else if (e instanceof Or)
					me.e = new And(lme.e, rme.e);
				else if (e instanceof ConditionalOr)
					me.e = new ConditionalAnd(lme.e, rme.e);
				else
					throw new MedorException(
						"BinaryLogicalOperator unmanaged by this method: " + e);
			}
			else {
				if (lme.isModified)
					((BinaryOperator) e).setExpression(0, lme.e);
				if (rme.isModified)
					((BinaryOperator) e).setExpression(1, rme.e);
				me.isModified = lme.isModified || rme.isModified;
			}
		}
		else if (hasANot
			&& e instanceof Operand
			&& e.getType().getTypeCode() == PType.TYPECODE_BOOLEAN) {
			me.e = new Not(e);
			me.isModified = true;
		}
		return me;
	}
}
