/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */

package org.objectweb.medor.optim.lib;

import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.api.BinaryOperator;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.expression.lib.And;
import org.objectweb.medor.expression.lib.ConditionalAnd;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.query.api.FilteredQueryTree;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.lib.Project;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.HashMap;
import java.util.Map;

/**
 * This class represents the rule to push selections down a QueryTree.
 */
public class PushSelectionRule extends BasicRule {

    public class SameQT {
        public boolean isSame;
        public QueryTree qt;

        public SameQT(boolean i, QueryTree q) {
            isSame = i;
            qt = q;
        }
        public SameQT() {
        }
    }

	public PushSelectionRule() {
		super("PushSelectionRule");
	}

    public QueryTree rewrite(QueryTree qt, QueryNode parent) throws MedorException {
		debug = log != null && log.isLoggable(BasicLevel.DEBUG);
        rewriteExp(qt, null);
        return qt;
    }

    protected void rewriteExp(QueryTree qt, Expression addedExp) throws MedorException {
        if (debug) log.log(BasicLevel.DEBUG, "****(qt=" + qt + ", addedExp="+ addedExp + ")");
        if (!(qt instanceof QueryLeaf)) {
			// treat the FielOperand of the AddedExp to link them to the queryChild
			addedExp = goDownFieldOperand(addedExp).e;
		}
        if (debug) log.log(BasicLevel.DEBUG, "addedExp="+ addedExp);

        FilteredQueryTree fqt = null;
        ModifiedExpression me = null;
        if (qt instanceof FilteredQueryTree && qt.getClass()!=Project.class) {
            fqt = (FilteredQueryTree) qt;
            //Merge the added exp with the exp of the QueryTree
            me = merge(fqt.getQueryFilter(), addedExp, 1);

            if (qt instanceof QueryLeaf) {
                if (me.isModified)
                    fqt.setQueryFilter(me.e);
                //Do not extract expression for a QueryLeaf because it 's leaf
                // and the extracted expression could be assign to children of the
                // leaf
                return;
            }
        }
        else {
            me = new ModifiedExpression();
            me.e = addedExp;
        }

        Map qt2e = new HashMap();
        me = extractMapFromExpression(me.e, qt2e, 0);
        if (debug) log.log(BasicLevel.DEBUG, "me.isModified="+ me.isModified);
        if (debug) log.log(BasicLevel.DEBUG, "fqt="+ fqt);
        if (me.isModified && fqt!=null) {
            // the expression has been modified, some sub expressions has bean
            // extracted
            fqt.setQueryFilter(me.e);
        }
        QueryTree[] qts = ((QueryNode) qt).getChildren();
        for (int i=0; i<qts.length; i++) {
            //Recall the function on the children, with the extracted sub
            // expression corresponding to the child
            rewriteExp(qts[i], (Expression) qt2e.get(qts[i]));
        }
    }

    /**
     * It merges an source expresion with an additional expression.
     * @param origin is the source expression
     * @param addExp is the additionnal expression which will be merge with the
     * source.
     * @return a ModifiedExpression instance. The 'isModified' field indicates
     * if the source expression has been modified ou replaced. The 'e' field
     * references the result of the merge.
     */
    public ModifiedExpression merge(Expression origin, Expression addExp, int op) {
        ModifiedExpression result = new ModifiedExpression();
        if (debug) log.log(BasicLevel.DEBUG, "Merge(" + origin + ", " + addExp + ", " + op + ")");
        if (addExp == null) {
            result.e = origin;
            result.isModified = false;
        } else if (origin == null) {
            result.e = addExp;
            result.isModified = true;
        } else {
            if (op == 2)
                result.e = new ConditionalAnd(origin, addExp);
            else
                result.e = new And(origin, addExp);
            result.isModified = true;
        }
        return result;
    }

    /**
     * This methods does two things. It extracts the expressions which can be
     * moved to a QueryTree child. The extracted expression are grouped by
     * QueryTree child in the Map parameter.
     * @param e is the expression in which sub expression could be extracted
     * @param qt2e is a map result parameter which must be fill with the
     * the extractable sub expressions.
     * @return a modifiedExpression instance. The 'isModified' field indicates
     * if the source expression has been modified ou replaced. The 'e' field
     * references the rest of the expression.
     */
    public ModifiedExpression extractMapFromExpression(
            Expression e, Map qt2e, int op) throws MedorException {
        ModifiedExpression result = new ModifiedExpression();
        if (debug) log.log(BasicLevel.DEBUG, "extractMapFromExpression(" + e + ")");
        if (e == null) {
            result.e = e;
            result.isModified = false;
        } else if (e instanceof And || e instanceof ConditionalAnd) {
            int curOp = (e instanceof And ? 1 : 2);
            ModifiedExpression l = extractMapFromExpression(
                    ((BinaryOperator) e).getExpression(0), qt2e, curOp);
            ModifiedExpression r = extractMapFromExpression(
                    ((BinaryOperator) e).getExpression(1), qt2e, curOp);

            if (l.e == null) {
                //The left expression has been extracted then used the right
                result.isModified = true;
                result.e = r.e;
            } else if (r.e == null) {
                //The right expression has been extracted then used the left
                result.isModified = true;
                result.e = l.e;
            } else {
                //r!=null && l!=null => the reference to the expression does
                // not change
                result.e = e;
                // By default we suppose that there is no change
                result.isModified = false;
                if (l.isModified) {
                    // The left has changed => assign the new expression
                    ((BinaryOperator) e).setExpression(0, l.e);
                    result.isModified = true;
                }
                if (r.isModified) {
                    // The right has changed => assign the new expression
                    ((BinaryOperator) e).setExpression(1, r.e);
                    result.isModified = true;
                    result.e = e;
                }
            }
        } else {
            // check if the expression is compose by fields which are linked to
            // to the same QueryTree
            SameQT sqt = isSameQT(e);
            if (sqt.isSame) {
                // Same QueryTree in the expression
                if (debug) log.log(BasicLevel.DEBUG, "extractMapFromExpression: same QT: sqt.qt=" + sqt.qt);
                ModifiedExpression me = merge((Expression) qt2e.get(sqt.qt), e, op);
                qt2e.put(sqt.qt, me.e);
                //The current expression can be moved to the queryTree child
                // then the expression returned is null
                result.isModified = true;
                result.e = null;
            } else {
                if (debug) log.log(BasicLevel.DEBUG, "extractMapFromExpression: not same QT:");
                //It is not the same QueryTree in the expression
                // then it is not possible to move this expression
                result.isModified = false;
                result.e = e;
            }
        }
        return result;
    }

    /**
     * It evaluates if inside an expression all QueryTreeField referenced in
     * FieldOperands come from the same QueryTree.
     * @param e is the expression in which the FieldOperand field will be
     * searched
     * @return a SameQt instance. The 'isSame' field indicates if in the
     * given expression all Fields are linked to the same QueryTree. If 'isSame'
     * is true, the 'qt' field is the QueryTree with which all fieldoperand of
     * expression  are linked.
     */
    public SameQT isSameQT(Expression e) throws MedorException {
        if (debug) log.log(BasicLevel.DEBUG, "isSameQT(" + e +")");
        if (e instanceof Operator) {
			Operator op = (Operator) e;
			SameQT res = new SameQT();
			res.isSame = true;
			SameQT c = null;
			for (int i=0; i<op.getOperandNumber() && res.isSame; i++) {
				c = isSameQT(((Operator) e).getExpression(i));
				if (debug) log.log(BasicLevel.DEBUG, "isSameQT(): c.isSame=" + c.isSame);
				if (debug) log.log(BasicLevel.DEBUG, "isSameQT(): c.qt=" + c.qt);
				res.isSame &= c.isSame;
				if (res.qt == null) {
					res.qt = c.qt;
				} else if (c.qt != null)
					res.isSame &= (c.qt == res.qt);
			}
			return res;
        } else if (e instanceof FieldOperand) {
            QueryTree qtChild = ((QueryTreeField) ((FieldOperand) e).getField())
                    .getQueryTree();
            return new SameQT(true, qtChild);
        } else {
            return new SameQT(true, null);
        }
    }

    public ModifiedExpression goDownFieldOperand(Expression e)
            throws MedorException {
        ModifiedExpression me = new ModifiedExpression();
        me.e = e;
        me.isModified = false;
        if (e instanceof Operator) {
			if (debug) log.log(BasicLevel.DEBUG, "Operator " + e);
			Operator op = (Operator) e;
			for (int i=0; i<op.getOperandNumber(); i++) {
				me = goDownFieldOperand(op.getExpression(i));
				if (me.isModified)
					op.setExpression(i, me.e);
			}
			// but the current expression does not change
			me.e = e;
			me.isModified = false;
        } else if (e instanceof FieldOperand) {
			if (debug) log.log(BasicLevel.DEBUG, "FieldOperand " + e);
            FieldOperand fo = (FieldOperand) e;
            QueryTreeField f = (QueryTreeField) fo.getField();
            if (f.getQueryTree() instanceof QueryLeaf) {
				if (debug) log.log(BasicLevel.DEBUG, "FieldOperand QueryLeaf");
                // On a QueryLeaf the FieldOperand references Field of the same
                // QueryTree and not a Field of a QueryTree child.
                //me.e = e;
                //me.isModified = false;
            } else if (f instanceof PropagatedField) {
				if (debug) log.log(BasicLevel.DEBUG, "FieldOperand PropagatedField " + f);
                //assign the field of child to the PropagatedField
                fo.setField(((PropagatedField) f).getPreviousFields()[0]);
                //me.e = e;
                //me.isModified = false;
            } else if (f instanceof CalculatedField) {
				if (debug) log.log(BasicLevel.DEBUG, "FieldOperand CalculatedField" + f);
                //replace the CalculatedField by its value
                me.e = ((CalculatedField) e).getExpression();
                me.isModified = true;
            } else
                throw new MedorException("Impossible move other type:" + f);
        }
        return me;
    }
}