/**
 * Copyright (C) 2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.medor.optim.lib;

import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.FilteredQueryTree;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.filter.api.FieldOperand;

/**
 * This class is an helper class for managing the replacement of a QueryTree in
 * a medor query represented by another QueryTree.
 *
 * @author S.Chassande-Barrioz
 */
public class QueryNodeReplacer {

	/**
	 * Replace a QueryTree by another. this is doing by replacing the usage of
	 * its field in all tree.
	 * @param old is the node to replace
	 * @param neo is the new node replacing the old
	 * @param root is the QueryTree from the search must start.
	 * @return if the root queryTree is the one searched, then the newer is
	 * returned, otherwise the root QueryTree is returned.
	 */
	public QueryTree replaceNode(QueryTree old, QueryTree neo, QueryTree root) {
		if (root == old) {
			return neo;
		}
		replaceNode(old.getTupleStructure().getFields(),
			neo.getTupleStructure().getFields(),
			root);
		return root;
	}

	/**
	 * Replace the references to old field by references to new fields in a
	 * QueryTree. The oldFields must not be attached to the current QueryTree.
	 * (Recursive method on QueryNode)
	 * @param oldFields the list of old field to replace
	 * @param neoFields the list of new field replacing the old
	 * @param current is the current QueryTree to treat
	 */
	private void replaceNode(Field[] oldFields, Field[] neoFields, QueryTree current) {
		if (current instanceof FilteredQueryTree) {
			replaceFields(oldFields, neoFields,
				((FilteredQueryTree) current).getQueryFilter());
			//Update children from
			((FilteredQueryTree) current).setQueryFilter(
				((FilteredQueryTree) current).getQueryFilter());
		}
		if (current instanceof QueryNode) {
			QueryNode qn = (QueryNode) current;
			Field[] currentFields = qn.getTupleStructure().getFields();
			for (int i = 0; i < currentFields.length; i++) {
				if (currentFields[i] instanceof PropagatedField) {
					PropagatedField pf = (PropagatedField) currentFields[i];
					Field[] previous = pf.getPreviousFields();
					boolean found = false;
					//Search on previous field
					for (int j = 0; j < previous.length; j++) {
						for (int k = 0; k < oldFields.length; k++) {
							if (previous[j] == oldFields[k]) {
								found = true;
								//found ==> replace
								previous[j] = neoFields[k];
								break;
							}
						}
					}
					if (found) {
						pf.replacePreviousField(previous);
					}
				} else if (currentFields[i] instanceof CalculatedField) {
					CalculatedField cf = (CalculatedField) currentFields[i];
					replaceFields(oldFields, neoFields, cf.getExpression());
				}
			}
			QueryTree[] children = qn.getChildren();
			for (int i = 0; i < children.length; i++) {
				replaceNode(oldFields, neoFields, children[i]);
			}
		}

	}

	/**
	 * Replace the references to old field by references to new fields in an
	 * expression (Recusrive method).
	 * @param oldFields the list of old field to replace
	 * @param neoFields the list of new field replacing the old
	 * @param exp is the current medor expression to treat.
	 */
	public void replaceFields(Field[] oldFields, Field[] neoFields, Expression exp) {
		if (exp instanceof Operator) {
			Operator op = (Operator) exp;
			for (int i = 0; i < op.getOperandNumber(); i++) {
				replaceFields(oldFields, neoFields, op.getExpression(i));
			}
		} else if (exp instanceof FieldOperand) {
			FieldOperand fo = (FieldOperand) exp;
			Field f = fo.getField();
			for (int i = 0; i < oldFields.length; i++) {
				if (f == oldFields[i]) {
					fo.setField(neoFields[i]);
					break; //stop the search
				}
			}
		}
	}

}
