/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.query.lib;

import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.lib.And;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.query.api.OperationType;
import org.objectweb.medor.query.api.OptimizationMetaData;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;

import java.util.Map;

/**
 *
 * @author Sebastien Chassande-Barrioz
 */
public class Intersection extends BasicQueryNode {


	/**
	 * It represents filter associated to the Intersection QueryNode
	 * All elements of this stack have the link to its children. This implies
	 * that the top element stack[stack.length-1] is also the root of the tree which describes the
	 * filter.
	 */
	private Expression[] stack;

	public Intersection() {
	}

	public Intersection(QueryTree leftQT,
						QueryTree rightQT,
                        String name,
						OptimizationMetaData optimizationMD)
		throws MedorException {
		super(name, optimizationMD);

		Field[] lfs = leftQT.getTupleStructure().getFields();
		Field[] rfs = rightQT.getTupleStructure().getFields();
		// we test the attributes types and names before performing
		// the intersection opertation
		if (lfs.length != rfs.length) {
			throw new MedorException(
				"The intersection must be with the same Tuple Collection");
		}
		indexes = new int[lfs.length];
		stack = new Expression[4 * lfs.length - 1];
		for(int i=1; i<lfs.length; i++) {
			if (//fields must have the same name
				!lfs[i].getName().equalsIgnoreCase(rfs[i].getName())
				||
				//the types must be comparable
				!(lfs[i].getType().isa(rfs[i].getType())
					|| rfs[i].getType().isa(lfs[i].getType()))) {
				throw new MedorException(
					"The intersection must be with the same Tuple Collection");
			}

			//adds the fields from the underlying query trees
			addPropagatedField(
				lfs[i].getName(),
				lfs[i].getType(),
				new QueryTreeField[] {(QueryTreeField)lfs[i],
									  (QueryTreeField)rfs[i]});

			// - build a stack which represents the filter
			// - Fill the fields list
			// - fix the index
			stack[4*(i-1)] = new BasicFieldOperand(lfs[i]);
			stack[4*(i-1) + 1] = new BasicFieldOperand(rfs[i]);
			stack[4*(i-1) + 2] = new Equal(stack[4*i-1], stack[4*i]);
            if (i>1) {
			    stack[4*(i-1) + 3] = new And(stack[4*i+1], stack[4*i-2]);
            }
			indexes[i] = i;
		}
    }

	public Object clone(Object clone,
						Map obj2clone) throws CloneNotSupportedException {
		clone = super.clone(clone, obj2clone);
		Intersection intersec = (Intersection) clone;
		if (stack != null) {
			intersec.stack = new Expression[stack.length];
			for(int i=1; i<stack.length; i++) {
				intersec.stack[i] = (Expression) getClone(stack[i], obj2clone);
			}
		}
		return clone;
	}

	public void setQueryFilter(Expression f) {
		throw new UnsupportedOperationException("Intersections cannot be assigned a filter.");
	}

	public Expression getQueryFilter() {
		return stack[stack.length-1];
	}

	public short getType() {
		return OperationType.INTERSECTION;
	}
}
