/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */

package org.objectweb.medor.query.rdb.api;

import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.ExpressionException;

import java.util.ArrayList;

/**
 * An RdbQueryLeaf represents a query leaf which encapsulates access to a
 * relational database.
 * <p>
 * A SQL query is associated to an RdbQueryLeaf.
 * @see org.objectweb.medor.query.rdb.api.RdbExpQueryLeaf
 * @see org.objectweb.medor.query.rdb.api.RdbStringQueryLeaf
 *
 * @author Sebastien Chassande-Barrioz
 */
public interface RdbQueryLeaf extends QueryLeaf {

	/**
	 * Returns the associated SQL query.
	 * <p>n the case of an RdbExpQueryLeaf, the query is computed.
	 * @return the SQL query associated to the RdbQueryLeaf, as a String.
	 * @throws MedorException in the case of RdbExpQueryLeaves if the
	 * SQL request cannot be computed from the associated Expression
	 */
	String getSqlRequest(ParameterOperand[] pos, ArrayList selectFields)
			throws MedorException, ExpressionException;

	/**
	 * Returns the associated SQL query.
	 * <p>In the case of an RdbExpQueryLeaf, the query is computed.
	 * @return the SQL query associated to the RdbQueryLeaf, as a String.
	 * @throws MedorException in the case of RdbExpQueryLeaves if the
	 * SQL request cannot be computed from the associated Expression
	 */
	String getSqlRequest(ParameterOperand[] pos)
			throws MedorException, ExpressionException;

	/**
	 * Builds the SELECT clause of the query (without "SELECT ").
	 * @param selectList is the start of the SELECT clause to which the list
	 * of qualified columns is appended.
	 * @param selectFields is the fields for the SELECT clause. Such fields can
	 * either be plain fields, or can be aggregate fields
	 * @param qualified indicates whether the field names should be qualified
	 * with the table name or not.
	 * @return the list of qualified columns for the SELECT clause
	 */
	String getSelectList(String selectList, ArrayList selectFields,
						 boolean qualified)
			throws MedorException;

	/**
	 * Is the current query leaf a subquery (in particular for the case of
	 * aggregates).
	 * @return true if the current RdbExpQueryLeaf is a subquery
	 */
	boolean isSubquery();

	/**
	 * Indicates whether the current RdbExpQueryLeaf is a subquery of another
	 * Rdb query leaf (in particular in the case of subqueries).
	 * <p>The default value is false.</p>
	 * @param subquery is true when the current RdbExpQueryLeaf is a subquery
	 * of another Rdb query leaf/node, false otherwise.
	 */
	void setIsSubquery(boolean subquery);
}
