/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre, S. Chassande-Barrioz
 */

package org.objectweb.medor.query.rdb.lib;

import org.objectweb.jorm.mapper.rdb.adapter.api.JoinedTable;
import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapter;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.QueryLeafException;
import org.objectweb.medor.datasource.api.DataStore;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.filter.api.AggregateOperator;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.lib.Count;
import org.objectweb.medor.filter.lib.ExpressionPrinter;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.query.api.OrderField;
import org.objectweb.medor.query.rdb.api.QualifiedTable;
import org.objectweb.medor.query.rdb.api.RdbExpField;
import org.objectweb.medor.query.rdb.api.RdbExpQueryLeaf;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * This class represents a QueryLeaf that maps onto a relational database store.
 * It implements the case where the relational query is provided as a set of
 * QualifiedTables and an Expression representing the filter.
 */
public class BasicRdbExpQueryLeaf extends BasicRdbQueryLeaf
		implements RdbExpQueryLeaf {

	/**
	 * The associated MEDOR Expression2WhereClause converter
	 */
	private MedorExpression2WhereClause exp2where;

	QualifiedTable[] from;
	RdbExpField groupByField;
	String rdbAdapterName = null;
	RdbAdapter rdbAdapter = null;
	List rootJoinedTables;

	public BasicRdbExpQueryLeaf() {
	}

	public BasicRdbExpQueryLeaf(DataStore ds, String nodeName) {
		super(nodeName, null, ds);
		exp2where = new MedorExpression2WhereClause();
	}

	/**
	 * Constructs a BasicRdbExpQueryLeaf from an Expression and a set of
	 * QualifiedTables.
	 * <p>The SQL request (string) will be calculated.
	 * @param ds is the DataStore associated to the QueryLeaf to be created
	 * @param from is the array of QualifiedTables associated to the QueryLeaf
	 * to be created
	 * @throws QueryLeafException if the QualifiedTables have all the same
	 * names, and in this case also the same aliases.
	 */
	public BasicRdbExpQueryLeaf(DataStore ds,
								QualifiedTable[] from, String nodeName)
			throws QueryLeafException {
		this(ds, nodeName);
		//all QualifiedTables should have different names, and if not,
		//different aliases
		//TODO throw an exception if all QualifiedTables have the same name
		//and the same alias name
		this.from = from;
	}


	public Object clone(Object clone,
						Map obj2clone) throws CloneNotSupportedException {
		clone = super.clone(clone, obj2clone);
		BasicRdbExpQueryLeaf ql = (BasicRdbExpQueryLeaf) clone;
		if (from != null) {
			ql.from = new QualifiedTable[from.length];
			for(int i=0; i<from.length; i++) {
				ql.from[i] = (QualifiedTable) getClone(from[i], obj2clone);
			}
		}
		ql.groupByField = (RdbExpField) getClone(groupByField, obj2clone);
		ql.rdbAdapterName = rdbAdapterName;
		ql.rdbAdapter = rdbAdapter;
		ql.exp2where = exp2where;
		if (rootJoinedTables != null) {
			ql.rootJoinedTables = new ArrayList();
			for(int i=0; i<rootJoinedTables.size(); i++) {
				JoinedTable jt = (JoinedTable) rootJoinedTables.get(i);
				if (jt == null) {
					ql.rootJoinedTables.add(null);
				} else {
					ql.rootJoinedTables.add(jt.clone(null, obj2clone));
				}
			}
		}
		return clone;
	}

	public void setRdbAdapterName(String an) {
		logger.log(BasicLevel.DEBUG, "Setting rdbAdapterName to " + an);
		this.rdbAdapterName = an;
	}

	public String getRdbAdapterName() {
		return rdbAdapterName;
	}

	public void setRdbAdapter(RdbAdapter adapter) {
		logger.log(BasicLevel.DEBUG, "Setting rdbAdapter to " + adapter);
		rdbAdapter = adapter;
	}

	public RdbAdapter getRdbAdapter() {
		return rdbAdapter;
	}

	public void setRootJoinedTables(List rootjts) {
		rootJoinedTables = rootjts;
	}

	/**
	 * Adds a Field to the RdbExpQueryLeaf
	 * @param name is the name of the Field to be added.
	 * @param type is the PTyme of the Field to be added.
	 * @param colName is the name of the column in the associated
	 * QualifiedTable.
	 * @param table is the QualifiedTable from which the Field is to be created.
	 * @return the created RdbExpField.
	 */
	public RdbExpField addRdbField(String name,
								   PType type,
								   String colName,
								   QualifiedTable table)
			throws QueryLeafException {
		//check that the parameter QualifiedTable is a valid QualifiedTable
		//in this RdbExpQueryLeaf
		QualifiedTable[] qts = this.getQualifiedTables();
		int i = 0;
		while (i < qts.length) {
			if (table == qts[i])
				break;
			else
				i++;
		}
		if (i == qts.length)
			throw new QueryLeafException("The parameter QualifiedTable is not a QualifiedTable of the current RdbExpQueryLeaf");
		else {
			RdbExpField f =
					new BasicRdbExpField(name, type, colName, table, this);
			name2field.put(f.getName(), f);
			fields.add(f);
			return f;
		}
	}

	public void addQualifiedTable(QualifiedTable qt) {
		if (from == null)
			from = new QualifiedTable[]{qt};
		else {
			QualifiedTable[] qts = new QualifiedTable[from.length + 1];
			System.arraycopy(from, 0, qts, 0, from.length);
			qts[from.length] = qt;
			from = qts;
		}
	}

	public void removeQualifiedTable(QualifiedTable qt) throws MedorException {
		if (from == null)
			throw new MedorException("QualifiedTable " + qt + " is not in leaf " + this);
		for (int i = 0; i < from.length; i++) {
			if (from[i] == qt) {
				if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
					logger.log(BasicLevel.DEBUG, "Found table " + qt.getTableName());
				}
				QualifiedTable[] qts = new QualifiedTable[from.length - 1];
				System.arraycopy(from, 0, qts, 0, i);
				System.arraycopy(from, i+1, qts, i, from.length -i - 1);
				from = qts;
				return;
			}
		}
		throw new MedorException("QualifiedTable " + qt + " is not in leaf " + this);
	}

	public boolean containsQualifiedTable(QualifiedTable qt) {
		if (from == null)
			return false;
		for (int i = 0; i < from.length; i++) {
			if (from[i] == qt)
				return true;
		}
		return false;
	}

	/**
	 * Adds an RdbExpField to the RdbExpQueryLeaf
	 * @param rdbField is the RdbField to be added.
	 */
	public RdbExpField addRdbField(RdbExpField rdbField)
			throws QueryLeafException {
		rdbField.setQueryLeaf(this);
		name2field.put(rdbField.getName(), rdbField);
		fields.add(rdbField);
		return rdbField;
	}

	public void removeRdbField(RdbExpField rdbField)
			throws QueryLeafException {
		name2field.remove(rdbField.getName());
		fields.remove(rdbField);
	}


	public String getSqlRequest(ParameterOperand[] pos)
			throws MedorException, ExpressionException {
		return getSqlRequest(pos, null);
	}

	public void resetSqlRequest() {
		query = null;
	}

	/**
	 * Builds and returns the SQL query as a String corresponding to the
	 * current RdbExpQueryLeaf.
	 * <p>
	 * The SQL query is computed from the provided RdbExpFields (SELECT clause),
	 * QualifiedTables (FROM clause) and Expression (WHERE clause).
	 * @return the SQL query as a String.
	 */
	public String getSqlRequest(ParameterOperand[] pos, ArrayList selectFields)
			throws MedorException, ExpressionException {
		boolean debug = logger != null && logger.isLoggable(BasicLevel.DEBUG);
		if (debug) {
			logger.log(BasicLevel.DEBUG, "Entering getSqlRequest for " + this);
		}
		if (selectFields == null) {
			if (debug) {
				logger.log(BasicLevel.DEBUG, "Setting selectFields (was null) to local fields");
			}
			selectFields = fields;
		}
		if (query != null) {
			if (debug) {
				logger.log(BasicLevel.DEBUG, "Query already computed " + query);
			}
			return query;
		}
		String where = (sqlFilter == null ? ""
					: parseExpression(sqlFilter, pos));
		if (debug) {
			logger.log(BasicLevel.DEBUG, "WHERE clause for " + this + " : " + where);
		}
		StringBuffer querySB = new StringBuffer();
		if (rootJoinedTables != null && rootJoinedTables.size() > 0) {
			if (debug) {
				logger.log(BasicLevel.DEBUG,
					"Query with rootJoinedTables (adapter is "
					+ rdbAdapterName + " - " + rdbAdapter + ")");
			}
			querySB.append(rdbAdapter.getQuery(
					getSelectList("", selectFields, true),
					rootJoinedTables,
					where));
		} else {
			if (debug) {
				logger.log(BasicLevel.DEBUG, "Query without rootJoinedTables");
			}
			querySB.append(getSelectList("SELECT ", selectFields, true));
			querySB.append(getFromList());
			if (where != null && where.length() >0) {
				querySB.append(" WHERE ");
				querySB.append(where);
			}
		}
		OrderField[] orderByFields = this.getOrderBy();
		if ((orderByFields != null) && (orderByFields.length > 0)) {
			querySB.append(" ORDER BY ");
			String sep = "";
			for (int cpt = 0; (cpt < orderByFields.length); cpt++) {
				querySB.append(sep);
				sep = ", ";
				querySB.append(getQualifiedFieldName(
								(RdbExpField) orderByFields[cpt].getField()));
				querySB.append(orderByFields[cpt].getDesc() ? " DESC" : " ASC");
			}
		}
		if (groupByField != null) {
			querySB.append(" GROUP BY ");
			querySB.append(getQualifiedFieldName(groupByField));
		}
		query = querySB.toString();
		//In the case of OrderBy, and for Oracle, embed the query inside
		//a select *
		if (orderByFields != null && orderByFields.length > 0 && rdbAdapter != null) {
			query = rdbAdapter.handleOrderBy(query);
		}
		return query;
	}

	/**
	 * Sets the filter associated to the RbdQueryLeaf.
	 * <p>This filter is
	 * represented by an Expression. Unlike the filter of a QueryNode, is
	 * applies to its own Fields, instead of the Fields of its children.
	 * @param exp is the Medor Expression representing the filter to be applied
	 * to the relational source tables.
	 */
	public void setQueryFilter(Expression exp) {
		this.sqlFilter = exp;
	}

	/**
	 * Retrieves the filter associated to the RbdQueryLeaf.
	 * @return the Medor Expression representing the filter to be applied to
	 * the relational source tables.
	 */
	public Expression getQueryFilter() {
		return this.sqlFilter;
	}

	/**
	 * Returns the associated QualifiedTables.
	 * @return the array of associated QualifiedTables.
	 */
	public QualifiedTable[] getQualifiedTables() {
		return this.from;
	}

	public void setGroupBy(RdbExpField groupBy) throws QueryLeafException {
		if (fields.contains(groupBy)) {
			groupByField = groupBy;
		} else
			throw new QueryLeafException("This RdbExpField does not exist in the TupleStructure of this RdbQueryLeaf");
	}

	public RdbExpField getGroupByField() {
		return groupByField;
	}

	/**
	 * Builds the SELECT clause of the query (without "SELECT ").
	 * @param selectList is the start of the SELECT clause to which the list
	 * of qualified columns is appended.
	 * @param selectFields is the fields for the SELECT clause. Such fields can
	 * either be plain fields, or can be aggregate fields
	 * @param qualified indicates whether the field names should be qualified
	 * with the table name or not.
	 * @return the list of qualified columns for the SELECT clause
	 */
	public String getSelectList(String selectList, ArrayList selectFields,
								boolean qualified) {
		logger.log(BasicLevel.DEBUG, "Computing select list");
		selectList += (getDistinct() ? "DISTINCT " : "");
		for (int i = 0; i < selectFields.size(); i++) {
			if (selectFields.get(i) instanceof RdbExpField) {
				logger.log(BasicLevel.DEBUG, "Found RdbExpField " + selectFields.get(i));
				if (qualified) {
					selectList +=
							getQualifiedFieldName(
									((RdbExpField) selectFields.get(i)));
				} else {
					selectList +=
							((RdbExpField) selectFields.get(i)).getColumnName();
				}
			} else {
				logger.log(BasicLevel.DEBUG, "Found CalculatedField " + selectFields.get(i));
				AggregateOperator op = (AggregateOperator)
						((CalculatedField) selectFields.get(i)).getExpression();
				selectList += op.getOperatorString();
				selectList += "(";
				if (op.getDistinct()) {
					selectList += "DISTINCT ";
				}
				if ((op instanceof Count) && (((Count) op).countAll())) {
					selectList += "*";
				} else {
					RdbExpField theField =
							(RdbExpField)
							((FieldOperand) (op.getExpression(0))).getField();
					if (qualified) {
						selectList +=
								getQualifiedFieldName(theField);
					} else {
						selectList += theField.getColumnName();
					}
				}
				selectList += ")";
			}
			if (i < selectFields.size() - 1)
				selectList += ", ";
		}
		return selectList;
	}

	/**
	 * Builds the qualified name of a RdbExpField.
	 * @return the qualified name of an RdbExpField.
	 * @param field is the RdbExpField for which the qualified name is built.
	 */
	protected static String getQualifiedFieldName(RdbExpField field) {
		String res = null;
		if (field.getTable().getAliasName() == null)
			res = field.getTable().getTableName();
		else
			res = field.getTable().getAliasName();
		return res + "." + field.getColumnName();
	}

	/**
	 * Builds the FROM clause of the query (without "FROM").
	 * @return the list of qualified table names for the FROM clause
	 * tables is appended.
	 */
	protected String getFromList() {
		StringBuffer sb = new StringBuffer();
		sb.append(" FROM ");
		for (int i = 0; i < from.length; i++) {
			sb.append(from[i].getTableName());
			if (from[i].getAliasName() != null) {
				if (rdbAdapter != null) {
					rdbAdapter.writeTableAlias(from[i].getAliasName(), sb);
				} else {
					sb.append(" as ");
					sb.append(from[i].getAliasName());
				}
			}
			if (i < from.length - 1)
				sb.append(", ");
		}
		logger.log(BasicLevel.DEBUG, "FROM clause: " + sb.toString());
		return sb.toString();
	}

	/**
	 * Computes a portion of SQL statement for the WHERE clause from a Medor
	 * Expression.
	 * @param exp is the Medor Expression from which to compute the portion of
	 * the SQL WHERE clause.
	 * @return the portion of the SQL WHERE clause corresponding to the
	 * Expression.
	 */
	private String parseExpression(Expression exp, ParameterOperand[] pos)
			throws ExpressionException {
		logger.log(BasicLevel.DEBUG, "Parsing " + ExpressionPrinter.e2str(exp));
		return exp2where.convertExp2WhereClause(exp, rdbAdapter, this, pos);
	}

}
